"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import copy

import numpy as np
import jax.numpy as jnp
from flax import struct
import chex


OBJECT_TO_INDEX = {
	"unseen": 0,
	"empty": 1,
	"wall": 2,
	"floor": 3,
	"door": 4,
	"key": 5,
	"ball": 6,
	"box": 7,
	"goal": 8,
	"lava": 9,
	"agent": 10,
}


COLORS = {
    'red'   : np.array([255, 0, 0]),
    'green' : np.array([0, 255, 0]),
    'blue'  : np.array([0, 0, 255]),
    'purple': np.array([112, 39, 195]),
    'yellow': np.array([255, 255, 0]),
    'grey'  : np.array([100, 100, 100])
}


COLOR_TO_INDEX = {
    'red'   : 0,
    'green' : 1,
    'blue'  : 2,
    'purple': 3,
    'yellow': 4,
    'grey'  : 5,
}


# Map of agent direction indices to vectors
DIR_TO_VEC = jnp.array([
	# Pointing right (positive X)
	(1, 0), # right
	(0, 1), # down
	(-1, 0), # left
	(0, -1), # up
], dtype=jnp.int8)


@struct.dataclass
class EnvInstance:
	agent_pos: chex.Array
	agent_dir_idx: int
	goal_pos: chex.Array
	wall_map: chex.Array


def make_maze_map(
	params,
	wall_map, 
	goal_pos, 
	agent_pos, 
	agent_dir_idx,
	pad_obs=False):
	# Expand maze map to H x W x C
	empty = jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
	wall = jnp.array([OBJECT_TO_INDEX['wall'], COLOR_TO_INDEX['grey'], 0], dtype=jnp.uint8)
	maze_map = jnp.array(jnp.expand_dims(wall_map, -1), dtype=jnp.uint8)
	maze_map = jnp.where(maze_map > 0, wall, empty)
	
	agent = jnp.array([OBJECT_TO_INDEX['agent'], COLOR_TO_INDEX['red'], agent_dir_idx], dtype=jnp.uint8)
	agent_x,agent_y = agent_pos
	maze_map = maze_map.at[agent_y,agent_x,:].set(agent)

	goal = jnp.array([OBJECT_TO_INDEX['goal'], COLOR_TO_INDEX['green'], 0], dtype=jnp.uint8)
	goal_x,goal_y = goal_pos
	maze_map = maze_map.at[goal_y,goal_x,:].set(goal)

	# Add observation padding
	if pad_obs:
		padding = params.agent_view_size-1
	else:
		padding = 1

	maze_map_padded = jnp.tile(wall.reshape((1,1,*empty.shape)), (maze_map.shape[0]+2*padding, maze_map.shape[1]+2*padding, 1))
	maze_map_padded = maze_map_padded.at[padding:-padding,padding:-padding,:].set(maze_map)

	# Add surrounding walls
	wall_start = padding-1 # start index for walls
	wall_end_y = maze_map_padded.shape[0] - wall_start - 1
	wall_end_x = maze_map_padded.shape[1] - wall_start - 1
	maze_map_padded = maze_map_padded.at[wall_start,wall_start:wall_end_x+1,:].set(wall) # top
	maze_map_padded = maze_map_padded.at[wall_end_y,wall_start:wall_end_x+1,:].set(wall) # bottom
	maze_map_padded = maze_map_padded.at[wall_start:wall_end_y+1,wall_start,:].set(wall) # left
	maze_map_padded = maze_map_padded.at[wall_start:wall_end_y+1,wall_end_x,:].set(wall) # right

	return maze_map_padded



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import time
import argparse
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

import minimax.envs as envs
from minimax.envs.maze.maze import Actions
from minimax.envs.viz.grid_viz import GridVisualizer 


def redraw(state, obs, extras):
	if extras['is_ued_maze']:
		env_instance = extras['env'].get_env_instance(None, state)
		maze_map = extras['render_env'].set_env_instance(env_instance)[1].maze_map
		extras['viz'].render(extras['params'], state, highlight=False, maze_map=maze_map)
	else:
		extras['viz'].render(extras['params'], state)
		if extras['obs_viz'] is not None:
			extras['obs_viz'].render_grid(np.asarray(obs['image']), k_rot90=3, agent_dir_idx=3)

def reset(key, env, extras):
	key, subkey = jax.random.split(extras['rng'])
	obs, state = extras['jit_reset'](subkey)

	extras['rng'] = key
	extras['obs'] = obs
	extras['state'] = state
	extras['n'] += 1

	if not extras['is_ued_maze']:
		metrics = env.get_env_metrics(state)
		print(metrics)
		extras['n_walls_total'] += metrics['n_walls']

	if not extras['is_ued_maze']:
		print(f"mean walls: {extras['n_walls_total']/extras['n']}", flush=True)

	redraw(state, obs, extras)

def step(env, action, extras):
	key, subkey = jax.random.split(extras['rng']) 
	obs, state, reward, done, info = env.step_env(subkey, extras['state'], action)
	extras['obs'] = obs
	extras['state'] = state
	# print(f"reward={reward}, agent_dir={obs['agent_dir']}")
	print(f"reward={reward}")

	if done or action == Actions.done:
		key, subkey = jax.random.split(subkey) 
		reset(subkey, env, extras)
	else:
		redraw(state, obs, extras)

	extras['rng'] = key


def key_handler(env, extras, event):
	print('pressed', event.key)

	if event.key == 'escape':
		window.close()
		return

	if event.key == 'backspace':
		extras['jit_reset']((env, extras))
		return

	if event.key == 'left':
		step(env, Actions.left, extras)
		return
	if event.key == 'right':
		step(env, Actions.right, extras)
		return
	if event.key == 'up':
		step(env, Actions.forward, extras)
		return

	# Spacebar
	if event.key == ' ':
		step(env, Actions.toggle, extras)
		return
	if event.key == 'a':
		step(env, Actions.pickup, extras)
		return
	if event.key == 'd':
		step(env, Actions.drop, extras)
		return

	if event.key == 'enter':
		step(env, Actions.done, extras)
		return


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument(
		"--env",
		type=str,
		help="Environment name",
		default="Maze"
	)
	parser.add_argument(
		"--seed",
		type=int,
		help="random seed to generate the environment with",
		default=90
	)
	parser.add_argument(
		'--render_agent_view',
		default=False,
		help="draw the agent sees (partially observable view)",
		action='store_true'
	)
	parser.add_argument(
		'--height',
		default=13,
		type=int,
		help="height",
	)
	parser.add_argument(
		'--width',
		default=13,
		type=int,
		help="width",
	)
	parser.add_argument(
		'--n_walls',
		default=10,
		type=int,
		help="Number of walls",
	)
	parser.add_argument(
		'--agent_view_size',
		default=5,
		type=int,
		help="Number of walls",
	)
	parser.add_argument(
		"--screenshot_path",
		type=str,
		default=None,
		help="maze.png",
	)
	args = parser.parse_args()

	kwargs = dict(
		height=args.height,
		width=args.width,
		n_walls=args.n_walls,
		agent_view_size=args.agent_view_size,
		see_through_walls=True,
		see_agent=True,
		normalize_obs=False,
		sample_n_walls=False,
		replace_wall_pos=False,
		max_episode_steps=250,
	)
	kwargs = {}
	env, params = envs.make(args.env, kwargs)
	params = env.params

	is_ued_maze = False
	render_env = None
	if args.env.startswith('UEDMaze'):
		is_ued_maze = True
		render_env, _ = envs.make('Maze', kwargs)

	viz = GridVisualizer()
	obs_viz = None
	if args.render_agent_view:
		obs_viz = GridVisualizer()

	with jax.disable_jit(False):
		jit_reset = jax.jit(env.reset_env, static_argnums=(1,))
		key = jax.random.PRNGKey(args.seed)
		key, subkey = jax.random.split(key)
		o0, s0 = jit_reset(subkey)
		if is_ued_maze:
			maze_map = render_env.set_env_instance(env.get_env_instance(None, s0))[1].maze_map
			viz.render(params, s0, highlight=False, maze_map=maze_map)
		else:
			viz.render(params, s0)
			if obs_viz is not None:
				obs_viz.render_grid(np.asarray(o0['image']), k_rot90=3, agent_dir_idx=3)

		key, subkey = jax.random.split(key)
		extras = {
			'rng': subkey,
			'state': s0,
			'obs': o0,
			'params':params,
			'viz': viz,
			'obs_viz': obs_viz,
			'jit_reset': jit_reset,
			'n_walls_total': 0,
			'n': 0,
			'env': env,
			'render_env': render_env,
			'is_ued_maze': is_ued_maze
		}

		if args.screenshot_path is not None:
			print('saving')
			viz.screenshot(args.screenshot_path)

		viz.window.reg_key_handler(partial(key_handler, env, extras))
		viz.show(block=True)




"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse


def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from collections import namedtuple

import numpy as np
import jax
import jax.numpy as jnp

import minimax.util.pytree as _tree_util
from .ued_scores import compute_episodic_stats


RolloutBatch = namedtuple(
	'RolloutBatch', (
	'obs', 
	'actions', 
	'rewards', 
	'dones', 
	'log_pis',
	'values', 
	'targets', 
	'advantages',
	'carry'
))


class RolloutStorage:
	def __init__(
		self,
		discount,
		gae_lambda,
		n_envs,
		n_eval,
		n_steps,
		action_space,
		obs_space,
		agent,
		n_agents=1):
		self.discount = discount
		self.gae_lambda = gae_lambda
		self.n_agents = n_agents
		self.n_steps = n_steps
		self.n_envs = n_envs
		self.n_evals = n_eval
		self.flat_batch_size = n_envs*self.n_evals
		self.action_space = action_space
		self.value_ensemble_size = agent.model.value_ensemble_size

		dummy_rng = jax.random.PRNGKey(0)
		self.empty_obs = \
			jax.jax.tree_util.tree_map(
				lambda x: jnp.empty((n_agents, n_steps, self.flat_batch_size) \
						+ x.shape, dtype=x.dtype),
				obs_space.sample(dummy_rng))

		self.empty_action = \
			jax.jax.tree_util.tree_map(
				lambda x: jnp.empty((n_agents, n_steps, self.flat_batch_size) \
						+ x.shape, dtype=x.dtype),
				action_space.sample(dummy_rng))

		if agent.is_recurrent:
			self.empty_carry = \
				agent.init_carry(
					dummy_rng, batch_dims=(n_agents, self.n_steps, self.flat_batch_size))
		else:
			self.empty_carry = None

		if agent.is_recurrent:
			self.append = jax.vmap(self._append_with_carry, in_axes=0)
		else:
			self.append = jax.vmap(self._append_without_carry, in_axes=0)
		self.get_batch = jax.vmap(self._get_batch)
		self.get_return_stats = jax.vmap(self._get_return_stats, in_axes=(0, None))

	@partial(jax.jit, static_argnums=0)
	def reset(self):
		"""
		Maintains a pytree of rollout transitions and metadata
		"""
		if self.empty_carry is None:
			carry_buffer = None
		else:
			carry_buffer = self.empty_carry

		value_batch_size = (self.flat_batch_size,)
		if self.value_ensemble_size > 1:
			value_batch_size += (self.value_ensemble_size,)

		return {
			"obs": self.empty_obs,
			"actions": self.empty_action,
			"rewards": jnp.empty(
				(self.n_agents, self.n_steps, self.flat_batch_size), dtype=jnp.float32
			),
			"dones": jnp.empty((self.n_agents, self.n_steps, self.flat_batch_size), dtype=jnp.uint8),
			"log_pis_old": jnp.empty(
				(self.n_agents, self.n_steps, self.flat_batch_size), dtype=jnp.float32
			),
			"values_old": jnp.empty(
				(self.n_agents, self.n_steps, *value_batch_size), dtype=jnp.float32
			),
			"carry": carry_buffer,
			"_t": jnp.zeros((self.n_agents,), dtype=jnp.uint32) # for vmap
		}

	@partial(jax.jit, static_argnums=0)
	def _append(self, buffer, obs, action, reward, done, log_pi, value, carry):
		if carry is not None:
			carry_buffer = _tree_util.pytree_set_array_at(buffer["carry"], buffer["_t"], carry)
		else:
			carry_buffer = None

		return {
			"obs": _tree_util.pytree_set_struct_at(buffer["obs"], buffer["_t"], obs),
			"actions": _tree_util.pytree_set_struct_at(buffer["actions"], buffer["_t"], action),
			"rewards": buffer["rewards"].at[buffer["_t"]].set(reward.squeeze()),
			"dones": buffer["dones"].at[buffer["_t"]].set(done.squeeze()),
			"log_pis_old": buffer["log_pis_old"].at[buffer["_t"]].set(log_pi),
			"values_old": buffer["values_old"].at[buffer["_t"]].set(value.squeeze()),
			"carry": carry_buffer,
			"_t": (buffer["_t"] + 1) % self.n_steps,
		}

	@partial(jax.jit, static_argnums=0)
	def _append_with_carry(self, buffer, obs, action, reward, done, log_pi, value, carry):
		return self._append(buffer, obs, action, reward, done, log_pi, value, carry)

	@partial(jax.jit, static_argnums=0)
	def _append_without_carry(self, buffer, obs, action, reward, done, log_pi, value):
		return self._append(buffer, obs, action, reward, done, log_pi, value, None)

	@partial(jax.jit, static_argnums=(0,))
	def _get_batch(self, buffer, last_value):
		_dones = buffer["dones"]
		rewards = buffer["rewards"]

		gae, target = self.compute_gae(
			value=buffer["values_old"],
			reward=rewards,
			done=_dones,
			last_value=last_value
		)

		# T x N x E x M --> N x T x EM if recurrent or N x TEM if not
		if self.empty_carry is not None:
			carry = buffer["carry"]
		else:
			carry = None

		batch_kwargs = dict(
			obs=buffer["obs"],
			actions=buffer["actions"],
			rewards=rewards,
			dones=_dones,
			log_pis=buffer["log_pis_old"],
			values=buffer["values_old"],
			targets=target,
			advantages=gae,
			carry=carry,
		)

		return RolloutBatch(**batch_kwargs)

	def compute_gae(self, value, reward, done, last_value):
		def _compute_gae(carry, step):
			(discount, gae_lambda, gae, value_next) = carry
			value, reward, done = step

			value_diff = discount*value_next*(1-done) - value
			delta = reward + value_diff

			gae = delta + discount*gae_lambda*(1-done) * gae

			return (discount, gae_lambda, gae, value), gae

		value, reward, done = jnp.flip(value, 0), jnp.flip(reward, 0), jnp.flip(done, 0)

		# Handle ensemble values, which have an extra ensemble dim at index -1
		if value.shape != done.shape:
			reward = jnp.expand_dims(reward, -1)
			done = jnp.expand_dims(done, -1)

		gae = jnp.zeros(value.shape[1:])
		_, advantages = jax.lax.scan(
			_compute_gae,
			(self.discount, self.gae_lambda, gae, last_value),
			(value, reward, done),
			length=len(reward)
		)
		advantages = jnp.flip(advantages, 0)
		targets = advantages + jnp.flip(value, 0)

		return advantages, targets

	def _get_return_stats(self, rollout, control_idxs=None):
		if control_idxs is not None:
			positive_signs = (control_idxs == 0)
			reward_signs = -1*(positive_signs.astype(jnp.float32) - (~positive_signs).astype(jnp.float32))
			rewards = rollout["rewards"]*reward_signs
		else:
			rewards = rollout["rewards"]

		pop_batch_shape = (self.n_steps, self.n_envs, self.n_evals)
		rewards = jnp.flip(rewards.reshape(*pop_batch_shape), 0)
		dones = jnp.flip(rollout["dones"].reshape(*pop_batch_shape), 0)

		return compute_episodic_stats(rewards, dones)

	def set_final_reward(self, rollout, reward):
		rollout["rewards"] = rollout["rewards"].at[:,-1,:].set(reward)

		return rollout



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse
import json
import os
import pathlib

import numpy as np

from minimax.util.dotdict import DefaultDotDict
import minimax.config.xpid_maker as xpid_maker


def get_wandb_config():
	wandb_config_path = os.path.join(os.path.abspath(os.getcwd()), 'config', 'wandb.json')
	if os.path.exists(wandb_config_path):
		with open(wandb_config_path, 'r') as config_file:
			config = json.load(config_file)
			if len(config) == 2:
				return {
					'wandb_base_url': config['base_url'],
					'wandb_api_key': config['api_key'],
				}

	return {}


def generate_train_cmds(
	cmd, params, num_trials=1, start_index=0, newlines=False, 
	xpid_generator=None, xpid_prefix='', 
	include_wandb_group=False,
	count_set=None):
	separator = ' \\\n' if newlines else ' '
	
	cmds = []

	if xpid_generator:
		params['xpid'] = xpid_generator(cmd, params, xpid_prefix)
		if include_wandb_group:
			params['wandb_group'] = params['xpid']

	start_seed = params['seed']

	for t in range(num_trials):
		params['seed'] = start_seed + t + start_index

		_cmd = [f'python -m minimax.{cmd}']

		trial_idx = t + start_index
		for k,v in params.items():
			if v is None:
				continue

			if k == 'xpid':
				v = f'{v}_{trial_idx}'

				assert len(v) < 256, f'{v} exceeds 256 characters!'

				if count_set is not None:
					count_set.add(v)

			if v == "*":
				v = f'"*"'

			_cmd.append(f'--{k}={v}')

		_cmd = separator.join(_cmd)

		cmds.append(_cmd)

	return cmds


def generate_all_params_for_grid(grid, defaults={}):
	def update_params_with_choices(prev_params, param, choices):
		updated_params = []
		for v in choices:
			for p in prev_params:
				updated = p.copy()
				updated[param] = v
				updated_params.append(updated)

		return updated_params

	all_params = [{}]
	for param, choices in grid.items():
		all_params = update_params_with_choices(all_params, param, choices)

	full_params = []
	for p in all_params:
		d = defaults.copy()
		d.update(p)
		full_params.append(d)

	return full_params


def parse_args():
	parser = argparse.ArgumentParser(description='Make commands')
	
	parser.add_argument(
		'--dir',
		type=str,
		default='config/configs/',
		help='Path to directory with .json configs')

	parser.add_argument(
		'--config', '-c',
		type=str,
		default=None,
		help='Name of .json config for hyperparameter search-grid')

	parser.add_argument(
		'--n_trials',
		type=int,
		default=1,
		help='Name of .json config for hyperparameter search-grid')

	parser.add_argument(
		'--start_index',
		default=0,
		type=int,
		help='Starting trial index of xpid runs')

	parser.add_argument(
		'--count',
		action='store_true',
		help='Print number of generated commands at the end of output.')

	parser.add_argument(
		"--checkpoint",
		action='store_true',
		help='Whether to start from checkpoint'
	)

	parser.add_argument(
		"--wandb_base_url",
		type=str,
		default=None,
		help='wandb base url'
	)
	parser.add_argument(
		"--wandb_api_key",
		type=str,
		default=None,
		help='wandb api key'
	)
	parser.add_argument(
		'--wandb_project',
		type=str,
		default=None,
		help='wandb project name')

	parser.add_argument(
		'--include_wandb_group',
		action="store_true",
		help='Whether to include wandb group in cmds.')

	return parser.parse_args()


def xpid_from_params(cmd, p, prefix=''):
	p = DefaultDotDict(p)

	env_info = xpid_maker.get_env_info(p)
	runner_info = xpid_maker.get_runner_info(p)
	a_algo_info = xpid_maker.get_algo_info(p, role='student')

	a_info = a_algo_info
	if cmd != 'finetune':
		a_model_info = xpid_maker.get_model_info(p, role='student')
		a_info = f"{a_info}_{a_model_info}"
		pt_info = ''
	else:
		pt_agent_info = 'tch' if p.get('ft_teacher') else 'st'
		pt_info = f"-{p.get('checkpoint_name', 'checkpoint')}_{pt_agent_info}"

	tch_info = ''
	train_runner = p.get('train_runner', 'dr')
	if train_runner == 'paired':
		tch_algo_info = xpid_maker.get_algo_info(p, role='teacher')
		tch_model_info = xpid_maker.get_model_info(p, role='teacher')
		tch_info = f"_tch_{tch_algo_info}_{tch_model_info}"

	xpid = f"{train_runner}-{env_info}-{runner_info}-{a_info}{tch_info}{pt_info}"

	return xpid


def setup_config_dir():
	config_dir = 'config/configs'
	if not os.path.exists(os.path.join(config_dir, 'maze')):
		os.makedirs(config_dir, exist_ok=True)

		import shutil

		this_path = os.path.dirname(os.path.abspath(__file__))
		src_path = os.path.join(this_path, 'configs')

		for item in os.listdir(src_path):
			src_item = os.path.join(src_path, item)
			dst_item = os.path.join(config_dir, item)

			if os.path.isdir(src_item):
				shutil.copytree(src_item, dst_item, symlinks=True)
			else:
				shutil.copy(src_item, dst_item)


if __name__ == '__main__':
	args = parse_args()

	# Default parameters
	params = {
		# Not needed.
	}

	setup_config_dir()

	json_filename = args.config
	if not json_filename.endswith('.json'):
		json_filename += '.json'

	grid_path = os.path.join(os.path.expandvars(os.path.expanduser(args.dir)), json_filename)
	config = json.load(open(grid_path))
	cmd = config.get('cmd', 'train')
	grid = config['args']
	xpid_prefix = '' if 'xpid_prefix' not in config else config['xpid_prefix']

	if args.checkpoint:
		params['checkpoint'] = True

	if 'wandb_project' in grid:	
		params['wandb_project'] = args.wandb_project
		
		if args.wandb_base_url:
			params['wandb_base_url'] = args.wandb_base_url
		if args.wandb_api_key:
			params['wandb_api_key'] = args.wandb_api_key

		params.update(get_wandb_config())

	# Generate all parameter combinations within grid, using defaults for fixed params
	all_params = generate_all_params_for_grid(grid, defaults=params)

	unique_xpids = None
	if args.count:
		unique_xpids = set()

	# Print all commands
	if cmd == 'eval':
		xpid_generator = None
	else:
		xpid_generator = xpid_from_params
	count = 0
	for p in all_params:
		cmds = generate_train_cmds(
			cmd, p,
			num_trials=args.n_trials, 
			start_index=args.start_index, 
			newlines=True, 
			xpid_generator=xpid_generator, 
			xpid_prefix=xpid_prefix,
			include_wandb_group=args.include_wandb_group,
			count_set=unique_xpids)

		for c in cmds:
			print(c + '\n')
			count += 1

	if args.count:
		print(f'Generated {len(unique_xpids)} unique commands.')
		print('Sweep over')
		grid_sizes = []
		for k,v in grid.items():
			if len(v) > 1:
				grid_sizes.append(len(v))
				print(f'{k}: {len(v)}')

		print(f'Total num settings: {np.prod(grid_sizes)}')




"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P
from flax import struct
import chex
import numpy as np

from .ued_scores import UEDScore


class PLRBuffer(struct.PyTreeNode):
	levels: chex.Array
	scores: chex.Array
	ages: chex.Array
	max_returns: chex.Array # for MaxMC
	filled: chex.Array
	filled_count: chex.Array
	n_mutations: chex.Array

	ued_score: int = struct.field(pytree_node=False, default=UEDScore.L1_VALUE_LOSS.value)
	replay_prob: float = struct.field(pytree_node=False, default=0.5)
	buffer_size: int = struct.field(pytree_node=False, default=100)
	staleness_coef: float = struct.field(pytree_node=False, default=0.3)
	temp: float = struct.field(pytree_node=False, default=1.0)
	use_score_ranks: bool = struct.field(pytree_node=False, default=True)
	min_fill_ratio: float = struct.field(pytree_node=False, default=0.5)
	use_robust_plr: bool = struct.field(pytree_node=False, default=False)
	use_parallel_eval: bool = struct.field(pytree_node=False, default=False)


class PLRManager:
	def __init__(
		self,
		example_level, # Example env instance
		ued_score,
		replay_prob=0.5,
		buffer_size=100,
		staleness_coef=0.3,
		temp=1.0,
		min_fill_ratio=0.5,
		use_score_ranks=True,
		use_robust_plr=False,
		use_parallel_eval=False,
		comparator_fn=None,
		n_devices=1):

		assert not (ued_score == UEDScore.MAX_MC and not use_score_ranks), \
			'Cannot use proportional normalization with MaxMC, which can produce negative scores.'

		self.ued_score = ued_score
		self.replay_prob = replay_prob
		self.buffer_size = buffer_size
		self.staleness_coef = staleness_coef
		self.temp = temp
		self.min_fill_ratio = min_fill_ratio
		self.use_score_ranks = use_score_ranks
		self.use_robust_plr = use_robust_plr
		self.use_parallel_eval = use_parallel_eval
		self.comparator_fn = comparator_fn

		self.n_devices = n_devices

		example_level = jax.tree_map(lambda x: jnp.array(x), example_level)
		self.levels = jax.tree_map(
			lambda x: (
				jnp.tile(jnp.zeros_like(x), (buffer_size,) + (1,)*(len(x.shape)-1))).reshape(buffer_size, *x.shape),
			example_level)

		self.scores = jnp.full(buffer_size, -jnp.inf)
		self.max_returns = jnp.full(buffer_size, -jnp.inf)
		self.ages = jnp.zeros(buffer_size, dtype=jnp.uint32)
		self.filled = jnp.zeros(buffer_size, dtype=jnp.bool_)
		self.filled_count = jnp.zeros((1,), dtype=jnp.int32)
		self.n_mutations = jnp.zeros(buffer_size, dtype=jnp.uint32)


	partial(jax.jit, static_argnums=(0,))
	def reset(self):
		return PLRBuffer(
			ued_score=self.ued_score.value,
			replay_prob=self.replay_prob,
			buffer_size=self.buffer_size,
			staleness_coef=self.staleness_coef,
			temp=self.temp,
			min_fill_ratio=self.min_fill_ratio,
			use_robust_plr=self.use_robust_plr,
			use_parallel_eval=self.use_parallel_eval,
			levels=self.levels,
			scores=self.scores,
			max_returns=self.max_returns,
			ages=self.ages,
			filled=self.filled,
			filled_count=self.filled_count,
			n_mutations=self.n_mutations)

	partial(jax.jit, static_argnums=(0,))
	def _get_replay_dist(self, scores, ages, filled):
		# Score dist
		if self.use_score_ranks:
			sorted_idx = jnp.argsort(-scores) # Top first
			scores = jnp.zeros(self.buffer_size, dtype=jnp.float32)\
					.at[sorted_idx]\
					.set(1.0/(1 + jnp.arange(self.buffer_size)))

		scores = scores*filled
		score_dist = scores**(1/self.temp)
		z = score_dist.sum()
		z = jnp.where(jnp.equal(z,0),1,z)
		score_dist = jax.lax.select(
			jnp.greater(z, 0),
			score_dist/z,
			filled*1. # Assign equal weight to all present levels
		)

		# Staleness dist
		staleness_scores = ages*filled
		_z = staleness_scores.sum()
		z = jnp.where(jnp.equal(_z,0),1,_z)
		staleness_dist = jax.lax.select(
			jnp.greater(_z, 0),
			staleness_scores/z,
			score_dist # If no solutions are stale, do not sample from staleness dist
		)

		# Replay dist
		replay_dist = (1-self.staleness_coef)*score_dist \
			+ self.staleness_coef*staleness_dist

		return replay_dist

	partial(jax.jit, static_argnums=(0,))
	def _get_next_insert_idx(self, plr_buffer):
		return jax.lax.cond(
			jnp.greater(plr_buffer.buffer_size, plr_buffer.filled_count[0]), 
			lambda *_: plr_buffer.filled_count[0], 
			lambda *_: jnp.argmin(self._get_replay_dist(plr_buffer.scores, plr_buffer.ages, plr_buffer.filled))
		)
 
	@partial(jax.jit, static_argnums=(0,3))
	def _sample_replay_levels(self, rng, plr_buffer, n):
		def _sample_replay_level(carry, step):
			ages = carry
			subrng = step
			replay_dist = self._get_replay_dist(plr_buffer.scores, ages, plr_buffer.filled)
			replay_idx = jax.random.choice(subrng, np.arange(self.buffer_size), shape=(), p=replay_dist)
			replay_level = jax.tree_map(lambda x: x.take(replay_idx, axis=0), plr_buffer.levels)

			ages = ((ages + 1)*(plr_buffer.filled)).at[replay_idx].set(0)

			return ages, (replay_level, replay_idx)

		rng, *subrngs = jax.random.split(rng, n+1)
		next_ages, (replay_levels, replay_idxs) = jax.lax.scan(
			_sample_replay_level,
			plr_buffer.ages,
			jnp.array(subrngs)
		)

		next_plr_buffer = plr_buffer.replace(
			ages=next_ages
		)

		return replay_levels, replay_idxs, next_plr_buffer

	def _sample_buffer_uniform(self, rng, plr_buffer, n):
		rand_idxs = jax.random.choice(rng, np.arange(self.buffer_size), shape=(n,), p=plr_buffer.filled)
		levels = jax.tree_map(lambda x: x.take(replay_idx, axis=0), plr_buffer.levels)

		return levels, rand_idxs, plr_buffer

	# Levels must be sampled sequentially, to account for staleness
	partial(jax.jit, static_argnums=(0,4,5))
	def sample(self, rng, plr_buffer, new_levels, n, random=False):
		rng, replay_rng, sample_rng = jax.random.split(rng, 3)

		is_replay = jnp.greater(self.replay_prob, jax.random.uniform(replay_rng))
		is_warm = jnp.greater_equal(plr_buffer.filled.sum()/self.buffer_size, self.min_fill_ratio)

		if self.n_devices > 1: # Synchronize replay
			is_replay = jax.lax.all_gather(is_replay, axis_name='device')[0]
			is_warm = jnp.all(jax.lax.all_gather(is_warm, axis_name='device'))

		is_replay = jnp.logical_and(is_replay, is_warm)

		if random:
			sample_fn = self._sample_buffer_uniform
		else:
			sample_fn = self._sample_replay_levels

		levels, level_idxs, next_plr_buffer = jax.lax.cond(
			is_replay,
			partial(sample_fn, n=n),
			lambda *_: (new_levels, np.full(n, -1), plr_buffer),
			*(sample_rng, plr_buffer)
		)

		# Update ages when not sampling replay
		next_plr_buffer = jax.lax.cond(
			is_replay,
			lambda *_: next_plr_buffer,
			lambda *_: next_plr_buffer.replace(ages=(plr_buffer.ages+n)*(plr_buffer.filled))
		)

		return levels, level_idxs, is_replay, next_plr_buffer

	@partial(jax.jit, static_argnums=(0,))
	def dedupe_levels(self, plr_buffer, levels, level_idxs):
		if self.comparator_fn is not None and level_idxs.shape[-1] > 2:
			def _check_equal(carry, step):
				match_idxs, other_levels, is_self  = carry
				batch_idx, level = step

				matches = jax.vmap(self.comparator_fn, in_axes=(0,None))(other_levels, level)

				top2match, top2match_idxs = jax.lax.top_k(matches,2)

				is_self_dupe = jnp.logical_and(is_self, top2match[1]) # More than 1 match
				is_dedupe_idx = jnp.logical_and(is_self_dupe, jnp.greater(batch_idx, top2match_idxs[0]))
				self_match_idx = top2match_idxs[0]*is_dedupe_idx - (~is_dedupe_idx)

				_match_idx = jnp.where(
					is_self,
					self_match_idx, # only first
					top2match_idxs[0], # use first matching index in buffer
				)

				match_idxs = jnp.where(
					matches.any(),
					match_idxs.at[batch_idx].set(_match_idx),
					match_idxs
				)

				return (match_idxs, other_levels, is_self), None

			# dedupe among batch levels
			batch_dupe_idxs = jnp.full_like(level_idxs, -1)
			(batch_dupe_idxs, _, _), _ = jax.lax.scan(
				_check_equal,
				(batch_dupe_idxs, levels, True),
				(np.arange(level_idxs.shape[-1]), levels)
			)
			batch_dupe_mask = jnp.greater(batch_dupe_idxs, -1)

			# dedupe against PLR buffer levels
			(level_idxs, _, _), _ = jax.lax.scan(
				_check_equal,
				(level_idxs, plr_buffer.levels, False),
				(np.arange(level_idxs.shape[-1]), levels)
			)

			return level_idxs, batch_dupe_mask
		else:
			return level_idxs, jnp.zeros_like(level_idxs, dtype=jnp.bool_)

	partial(jax.jit, static_argnums=(0,7))	
	def update(self, plr_buffer, levels, level_idxs, ued_scores, dupe_mask=None, info=None, ignore_val=-jnp.inf, parent_idxs=None):
		# Note: parent_idxs are only used for mutated levels
		done_masks = (ued_scores != ignore_val)
		if dupe_mask is not None:
			done_masks = jnp.logical_and(done_masks, ~dupe_mask) # Ignore duplicate levels in batch by treating them as not done

		cur_n_mutations = plr_buffer.n_mutations
		insert_mask = jnp.zeros((self.buffer_size,), dtype=jnp.bool_)
		
		def update_level_info(carry, step):
			plr_buffer, insert_mask = carry
			levels = plr_buffer.levels
			scores = plr_buffer.scores
			filled = plr_buffer.filled

			score, level, level_idx, done_mask, parent_idx, max_return = step

			next_insert_idx = self._get_next_insert_idx(plr_buffer)
			is_new_level = jnp.greater(0, level_idx)
			insert_idx = jnp.where(
				is_new_level,
				next_insert_idx, # new level
				level_idx,
			)

			should_insert = jnp.greater_equal(score, scores.at[insert_idx].get())
			should_insert = jnp.logical_and(should_insert, done_mask)

			is_existing_level = jnp.logical_and(~is_new_level, done_mask)
			should_update = jnp.logical_and(is_existing_level, ~insert_mask.at[insert_idx].get())
			should_insert = jnp.logical_and(should_insert, ~should_update)
			next_insert_mask = jnp.where(
				should_insert,
				insert_mask.at[insert_idx].set(True),
				insert_mask
			)
			should_insert_or_update = jnp.logical_or(should_insert, should_update)

			# Update max return if needed
			next_max_returns = jnp.where(
				should_insert_or_update,
				plr_buffer.max_returns.at[insert_idx].set(max_return),
				plr_buffer.max_returns
			)

			updated_level = jax.tree_map(
				lambda x,y: jax.lax.select(should_insert, x, y),
				level,
				jax.tree_map(lambda x: x.at[insert_idx].get(), levels)
			)
			next_levels = jax.tree_map(lambda x,y: x.at[insert_idx].set(y), levels, updated_level)

			next_scores = jnp.where(
				should_insert_or_update,
				scores.at[insert_idx].set(score),
				scores
			)
			next_filled = jnp.where(
				should_insert,
				filled.at[insert_idx].set(True),
				filled
			)

			plr_replace_kwargs = dict(
				levels=next_levels,
				scores=next_scores,
				filled=next_filled,
				filled_count=jnp.array([next_filled.sum()]),
				max_returns=next_max_returns
			)

			# Update mutation count
			n_mutations = plr_buffer.n_mutations
			should_incr_n_mutations = jnp.logical_and(jnp.not_equal(parent_idx,-1), should_insert)
			should_reset_n_mutations = jnp.logical_and(jnp.equal(parent_idx,-1), should_insert_or_update)
			reset_n_mutations = jnp.where(
				is_existing_level,
				cur_n_mutations.at[insert_idx].get(),
				0
			)
			next_n_mutations = jnp.where(
				should_incr_n_mutations,
				n_mutations.at[insert_idx].set(cur_n_mutations.at[parent_idx].get() + 1),
				n_mutations
			)
			next_n_mutations = jnp.where(
				should_reset_n_mutations,
				n_mutations.at[insert_idx].set(reset_n_mutations),
				next_n_mutations
			)
			
			plr_replace_kwargs['n_mutations'] = next_n_mutations

			next_plr_buffer = plr_buffer.replace(**plr_replace_kwargs)

			return (next_plr_buffer, next_insert_mask), None

		if parent_idxs is None:
			parent_idxs = jnp.full_like(level_idxs, -1)

		if plr_buffer.ued_score == UEDScore.MAX_MC.value:
			max_returns = info['max_returns']
		else:
			max_returns = jnp.full_like(level_idxs, -1)
		carry = (ued_scores, levels, level_idxs, done_masks, parent_idxs, max_returns)

		(next_plr_buffer, _), _ = jax.lax.scan(
			update_level_info,
			(plr_buffer, insert_mask),
			carry
		)

		return next_plr_buffer

	@partial(jax.jit, static_argnums=(0,))
	def get_metrics(self, plr_buffer):
		replay_dist = self._get_replay_dist(plr_buffer.scores, plr_buffer.ages, plr_buffer.filled)
		weighted_n_mutations = (plr_buffer.n_mutations*replay_dist).sum()
		scores = jnp.where(plr_buffer.filled, plr_buffer.scores, 0)
		weighted_ued_score = (scores*replay_dist).sum()

		weighted_age = (plr_buffer.ages*replay_dist).sum()

		return dict(
			weighted_n_mutations=weighted_n_mutations,
			weighted_ued_score=weighted_ued_score,
			weighted_age=weighted_age
		)


class PopPLRManager(PLRManager):
	def __init__(self, *, n_agents, **kwargs):
		super().__init__(**kwargs)

		self.n_agents = n_agents

	@partial(jax.jit, static_argnums=(0,1))
	def reset(self, n):
		sup = super()
		return jax.vmap(lambda *_: sup.reset())(np.arange(n))

	partial(jax.jit, static_argnums=(0,4,5))
	def sample(self, rng, plr_buffer, new_levels, n, random=False):
		sup = super()
		
		rng, *vrngs = jax.random.split(rng, self.n_agents+1)

		return jax.vmap(sup.sample, in_axes=(0,0,0,None,None))(
			jnp.array(vrngs), plr_buffer, new_levels, n, random
		)

	@partial(jax.jit, static_argnums=(0,))
	def dedupe_levels(self, plr_buffer, levels, level_idxs):
		sup = super()
		return jax.vmap(sup.dedupe_levels)(plr_buffer, levels, level_idxs)

	partial(jax.jit, static_argnums=(0,7))
	def update(self, plr_buffer, levels, level_idxs, ued_scores, dupe_mask=None, info=None, ignore_val=-jnp.inf, parent_idxs=None):
		sup = super()
		return jax.vmap(sup.update, in_axes=(0,0,0,0,0,0,None,0))(
			plr_buffer, levels, level_idxs, ued_scores, dupe_mask, info, ignore_val, parent_idxs
		)

	partial(jax.jit, static_argnums=(0,))
	def get_metrics(self, plr_buffer):
		sup = super()
		return jax.vmap(sup.get_metrics)(plr_buffer)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file extends the Environment class from
https://github.com/RobertTLange/gymnax/

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import jax
import chex
from typing import Tuple, Union, Optional
from functools import partial
from flax import struct


@struct.dataclass
class EnvState:
	time: int


@struct.dataclass
class EnvParams:
	max_episode_steps: int


class Environment(object):
	"""Jittable abstract base class for all basic environments."""
	def __init__(self):
		self.eval_solved_rate = self.get_eval_solved_rate_fn()

	@property
	def default_params(self) -> EnvParams:
		return EnvParams()

	@staticmethod
	def align_kwargs(kwargs, other_kwargs):
		"""
		Return kwargs that are consistent with other_kwargs,
		e.g. in the case of the student env, other_kwargs may be 
		those for the paired teacher env, and in the case of the 
		teacher env, the paired student env.
		"""
		raise NotImplementedError

	@partial(jax.jit, static_argnums=(0,4))
	def step(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
		reset_on_done: bool = True,
		reset_state: Optional[chex.ArrayTree] = None,
	) -> Tuple[chex.ArrayTree, EnvState, float, bool]:
		"""Performs step transitions in the environment."""
		# Use default env parameters if no others specified
		if hasattr(self, 'params'):
			params = self.params
		else:
			params = self.default_params

		key, key_reset = jax.random.split(key)
		obs_st, state_st, reward, done, info = self.step_env(
			key, state, action
		)

		if reset_on_done:
			if reset_state is not None:
				state_re = reset_state
				obs_re = self.get_obs(reset_state)
			else:
				if hasattr(params, 'singleton_seed') \
					and params.singleton_seed >= 0:
					key_reset = jax.random.PRNGKey(params.singleton_seed)

				obs_re, state_re = self.reset_env(key_reset)

			# Auto-reset environment based on termination
			state = jax.tree_map(
				lambda x, y: jax.lax.select(done, x, y), state_re, state_st
			)
			obs = jax.tree_map(
				lambda x, y: jax.lax.select(done, x, y), obs_re, obs_st
			)
		else:
			obs, state = obs_st, state_st

		return obs, state, reward, done, info

	@partial(jax.jit, static_argnums=(0,))
	def reset(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""Performs resetting of environment."""
		# Use default env parameters if no others specified
		if hasattr(self, 'params'):
			params = self.params
		else:
			params = self.default_params

		if hasattr(params, 'singleton_seed') \
			and params.singleton_seed >= 0:
			key = jax.random.PRNGKey(params.singleton_seed)
		obs, state = self.reset_env(key)
		return obs, state

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
	) -> Tuple[chex.ArrayTree, EnvState, float, bool, dict]:
		"""Environment-specific step transition."""
		raise NotImplementedError

	def reset_env(
		self, key: chex.PRNGKey
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""Environment-specific reset."""
		raise NotImplementedError

	def set_state(
		self,
		state: EnvState
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""
		Implemented for basic envs.
		"""
		return self.get_obs(state), state

	def set_env_instance(
		self,
		encoding: chex.ArrayTree
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""
		Implemented for basic envs.
		"""
		raise NotImplementedError

	def get_env_instance(
		self,
		key: chex.PRNGKey,
		state: EnvState
	) -> chex.ArrayTree:
		"""
		Implemented for UED envs.
		"""
		raise NotImplementedError

	def get_obs(self, state: EnvState) -> chex.ArrayTree:
		"""Applies observation function to state."""
		raise NotImplementedError

	def is_terminal(self, state: EnvState) -> bool:
		"""Check whether state is terminal."""
		raise NotImplementedError

	def get_eval_solved_rate_fn(self):
		return None

	@property
	def name(self) -> str:
		"""Environment name."""
		return type(self).__name__

	@property
	def num_actions(self) -> int:
		"""Number of actions possible in environment."""
		raise NotImplementedError

	def action_space(self):
		"""Action space of the environment."""
		raise NotImplementedError

	def observation_space(self):
		"""Observation space of the environment."""
		raise NotImplementedError

	def state_space(self):
		"""State space of the environment."""
		raise NotImplementedError

	def max_episode_steps(self):
		"""Maximum number of time steps in environment."""
		raise NotImplementedError

	def get_env_metrics(self, state: EnvState):
		"""Environment-specific metrics, e.g. number of walls."""
		raise NotImplementedError



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import os
import copy

# os.environ['XLA_FLAGS'] = '--xla_force_host_platform_device_count=2'
# os.environ['JAX_TRACEBACK_FILTERING'] = 'off'
os.environ['XLA_PYTHON_CLIENT_MEM_FRACTION'] = '.70'

import jax
import wandb

from minimax.runners import ExperimentRunner
from minimax.util.loggers import Logger
from .arguments import parser


if __name__ == '__main__':
	with jax.disable_jit(False):
		args = parser.parse_args(preview=True)

		# === Setup the main runner ===
		_args = copy.deepcopy(args) # Mutable record of args
		xp_runner = ExperimentRunner(
			train_runner=_args.train_runner,
			env_name=_args.env_name,
			agent_rl_algo=_args.agent_rl_algo,
			student_model_name=_args.student_model_name,
			teacher_model_name=_args.teacher_model_name,
			train_runner_kwargs=_args.train_runner_args,
			env_kwargs=_args.env_args,
			ued_env_kwargs=_args.ued_env_args,
			student_rl_kwargs=_args.student_rl_args,
			teacher_rl_kwargs=_args.teacher_rl_args,
			student_model_kwargs=_args.student_model_args,
			teacher_model_kwargs=_args.teacher_model_args,
			eval_kwargs=_args.eval_args,
			eval_env_kwargs=_args.eval_env_args,
			n_devices=_args.n_devices,
		)
		
		# === Configure logging ===
		# Set up wandb
		wandb_args = args.wandb_args
		if wandb_args.base_url:
			os.environ["WANDB_BASE_URL"] = wandb_args.base_url
		if wandb_args.api_key:
			os.environ["WANDB_API_KEY"] = wandb_args.api_key
		if wandb_args.base_url and wandb_args.api_key:
			os.environ["WANDB_CACHE_DIR"] = '~/.cache/wandb'
			wandb.init(project=wandb_args.project, 
					   entity=wandb_args.entity, 
					   config=args, 
					   name=args.xpid, 
					   group=wandb_args.group)
			callback = wandb.log
		else:
			callback = None

		logger = Logger(
			log_dir=args.log_dir,
			xpid=args.xpid,
			xp_args=args,
			callback=callback,
			from_last_checkpoint=args.from_last_checkpoint,
			verbose=args.verbose)

		# === Start training ===
		rng = jax.random.PRNGKey(args.seed)
		xp_runner.train(
			rng=rng,
			n_total_updates=args.n_total_updates,
			logger=logger,
			log_interval=args.log_interval,
			test_interval=args.test_interval,
			checkpoint_interval=args.checkpoint_interval,
			archive_interval=args.archive_interval,
			archive_init_checkpoint=args.archive_init_checkpoint,
			from_last_checkpoint=args.from_last_checkpoint)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from enum import Enum
from functools import partial
from typing import Tuple, Optional
import inspect

import numpy as np
import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P
import optax
import flax
import flax.linen as nn
from flax.core.frozen_dict import FrozenDict

import minimax.envs as envs
from minimax.util import pytree as _tree_util
from minimax.util.rl import (
	AgentPop,
	VmapTrainState,
	RolloutStorage,
	RollingStats,
	UEDScore,
	compute_ued_scores
)


class PAIREDRunner:
	"""
	Orchestrates rollouts across one or more students and teachers. 
	The main components at play:
	- AgentPop: Manages train state and batched inference logic 
		for a population of agents.
	- BatchUEDEnv: Manages environment step and reset logic for a 
		population of agents batched over a pair of student and 
		teacher MDPs.
	- RolloutStorage: Manages the storing and sampling of collected txns.
	- PPO: Handles PPO updates, which take a train state + batch of txns.
	"""
	def __init__(
		self, 
		env_name,
		env_kwargs,
		ued_env_kwargs,
		student_agents,
		n_students=2,
		n_parallel=1,
		n_eval=1,
		n_rollout_steps=250,
		lr=1e-4,
		lr_final=None,
		lr_anneal_steps=0,
		max_grad_norm=0.5,
		discount=0.99,
		gae_lambda=0.95,
		adam_eps=1e-5,
		teacher_lr=None,
		teacher_lr_final=None,
		teacher_lr_anneal_steps=None,
		teacher_discount=0.99,
		teacher_gae_lambda=0.95,
		teacher_agents=None,
		ued_score='relative_regret',
		track_env_metrics=False,
		n_unroll_rollout=1,
		render=False,
		n_devices=1):
		assert n_parallel % n_devices == 0, 'Num envs must be divisible by num devices.'

		ued_score = UEDScore[ued_score.upper()]

		assert len(student_agents) == 1, \
			'Only one type of student supported.'
		assert not (n_students > 2 and ued_score in [UEDScore.RELATIVE_REGRET, UEDScore.MEAN_RELATIVE_REGRET]), \
			'Standard PAIRED uses only 2 students.'
		assert teacher_agents is None or len(teacher_agents) == 1, \
			'Only one type of teacher supported.'

		self.n_students = n_students
		self.n_parallel = n_parallel // n_devices
		self.n_eval = n_eval
		self.n_devices = n_devices
		self.step_batch_size = n_students*n_eval*n_parallel
		self.n_rollout_steps = n_rollout_steps
		self.n_updates = 0
		self.lr = lr
		self.lr_final = lr if lr_final is None else lr_final
		self.lr_anneal_steps = lr_anneal_steps
		self.teacher_lr = \
			lr if teacher_lr is None else lr
		self.teacher_lr_final = \
			self.lr_final if teacher_lr_final is None else teacher_lr_final
		self.teacher_lr_anneal_steps = \
			lr_anneal_steps if teacher_lr_anneal_steps is None else teacher_lr_anneal_steps
		self.max_grad_norm = max_grad_norm
		self.adam_eps = adam_eps
		self.ued_score = ued_score
		self.track_env_metrics = track_env_metrics

		self.n_unroll_rollout = n_unroll_rollout
		self.render = render

		self.student_pop = AgentPop(student_agents[0], n_agents=n_students)

		if teacher_agents is not None:
			self.teacher_pop = AgentPop(teacher_agents[0], n_agents=1)

		# This ensures correct partial-episodic bootstrapping by avoiding
		# any termination purely due to timeouts.
		# env_kwargs.max_episode_steps = self.n_rollout_steps + 1
		self.benv = envs.BatchUEDEnv(
			env_name=env_name,
			n_parallel=self.n_parallel,
			n_eval=n_eval,
			env_kwargs=env_kwargs,
			ued_env_kwargs=ued_env_kwargs,
			wrappers=['monitor_return', 'monitor_ep_metrics'],
			ued_wrappers=[]
		)
		self.teacher_n_rollout_steps = \
			self.benv.env.ued_max_episode_steps()

		self.student_rollout = RolloutStorage(
			discount=discount,
			gae_lambda=gae_lambda,
			n_steps=n_rollout_steps,
			n_agents=n_students,
			n_envs=self.n_parallel,
			n_eval=self.n_eval,
			action_space=self.benv.env.action_space(),
			obs_space=self.benv.env.observation_space(),
			agent=self.student_pop.agent
		)

		self.teacher_rollout = RolloutStorage(
			discount=teacher_discount,
			gae_lambda=teacher_gae_lambda,
			n_steps=self.teacher_n_rollout_steps,
			n_agents=1,
			n_envs=self.n_parallel,
			n_eval=1,
			action_space=self.benv.env.ued_action_space(),
			obs_space=self.benv.env.ued_observation_space(),
			agent=self.teacher_pop.agent,
		)

		ued_monitored_metrics = ('return',)
		self.ued_rolling_stats = RollingStats(	
			names=ued_monitored_metrics,
			window=10,
		)
		
		monitored_metrics = self.benv.env.get_monitored_metrics()
		self.rolling_stats = RollingStats(
			names=monitored_metrics,
			window=10,
		)

		self._update_ep_stats = jax.vmap(jax.vmap(self.rolling_stats.update_stats))
		self._update_ued_ep_stats = jax.vmap(jax.vmap(self.ued_rolling_stats.update_stats))

		if self.render:
			from envs.viz.grid_viz import GridVisualizer
			self.viz = GridVisualizer()
			self.viz.show()

	def reset(self, rng):
		self.n_updates = 0

		n_parallel = self.n_parallel*self.n_devices

		rng, student_rng, teacher_rng = jax.random.split(rng,3)
		student_info = self._reset_pop(
				student_rng, 
				self.student_pop, 
				partial(self.benv.reset, sub_batch_size=n_parallel*self.n_eval),
				n_parallel_ep=n_parallel*self.n_eval,
				lr_init=self.lr,
				lr_final=self.lr_final,
				lr_anneal_steps=self.lr_anneal_steps)

		teacher_info = self._reset_pop(
			teacher_rng, 
			self.teacher_pop, 
			partial(self.benv.reset_teacher, n_parallel=n_parallel),
			n_parallel_ep=n_parallel,
			lr_init=self.teacher_lr,
			lr_final=self.teacher_lr_final,
			lr_anneal_steps=self.teacher_lr_anneal_steps)

		return (
			rng,
			*student_info,
			*teacher_info
		)

	def _reset_pop(
		self, 
		rng, 
		pop, 
		env_reset_fn, 
		n_parallel_ep=1,
		lr_init=3e-4,
		lr_final=3e-4,
		lr_anneal_steps=0):
		rng, *vrngs = jax.random.split(rng, pop.n_agents+1)
		reset_out = env_reset_fn(jnp.array(vrngs))
		if len(reset_out) == 2:
			obs, state = reset_out
		else:
			obs, state, extra = reset_out
		dummy_obs = jax.tree_util.tree_map(lambda x: x[0], obs) # for one agent only

		rng, subrng = jax.random.split(rng)
		if pop.agent.is_recurrent:
			carry = pop.init_carry(subrng, obs)
		else:
			carry = None

		rng, subrng = jax.random.split(rng)
		params = pop.init_params(subrng, dummy_obs)

		schedule_fn = optax.linear_schedule(
			init_value=-float(lr_init),
			end_value=-float(lr_final),
			transition_steps=lr_anneal_steps,
		)

		tx = optax.chain(
			optax.clip_by_global_norm(self.max_grad_norm),
			optax.scale_by_adam(eps=self.adam_eps),
			optax.scale_by_schedule(schedule_fn),
		)

		train_state = VmapTrainState.create(
			apply_fn=pop.agent.evaluate,
			params=params,
			tx=tx
		)
		
		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=(pop.n_agents,n_parallel_ep))

		return train_state, state, obs, carry, ep_stats

	def get_checkpoint_state(self, state):
		_state = list(state)
		_state[1] = state[1].state_dict
		_state[6] = state[6].state_dict

		return _state

	def load_checkpoint_state(self, runner_state, state):
		runner_state = list(runner_state)
		runner_state[1] = runner_state[1].load_state_dict(state[1])
		runner_state[6] = runner_state[6].load_state_dict(state[6])

		return tuple(runner_state)

	@partial(jax.jit, static_argnums=(0,2,3))
	def _get_transition(
		self,
		rng, 
		pop, 
		rollout_mgr, 
		rollout, 
		params, 
		state, 
		obs, 
		carry, 
		done,
		reset_state=None,
		extra=None):
		# Sample action
		value, pi_params, next_carry = pop.act(params, obs, carry, done)
		pi = pop.get_action_dist(pi_params)
		rng, subrng = jax.random.split(rng)
		action = pi.sample(seed=subrng)
		log_pi = pi.log_prob(action)

		rng, *vrngs = jax.random.split(rng, pop.n_agents+1)

		if pop is self.student_pop:
			step_fn = self.benv.step_student
		else:
			step_fn = self.benv.step_teacher
		step_args = (jnp.array(vrngs), state, action)

		if reset_state is not None: # Needed for student to reset to same instance
			step_args += (reset_state,)

		if extra is not None:
			step_args += (extra,)
			next_obs, next_state, reward, done, info, extra = step_fn(*step_args)
		else:
			next_obs, next_state, reward, done, info = step_fn(*step_args)

		# Add transition to storage
		step = (obs, action, reward, done, log_pi, value)
		if carry is not None:
			step += (carry,)

		rollout = rollout_mgr.append(rollout, *step)

		if self.render and pop is self.student_pop:
			self.viz.render(
				self.benv.env.env.params, 
				jax.tree_util.tree_map(lambda x: x[0][0], state))

		return rollout, next_state, next_obs, next_carry, done, info, extra

	@partial(jax.jit, static_argnums=(0,2,3,4))
	def _rollout(
		self, 
		rng, 
		pop, 
		rollout_mgr,
		n_steps,
		params, 
		state, 
		obs, 
		carry, 
		done,
		reset_state=None, 
		extra=None, 
		ep_stats=None):
		rngs = jax.random.split(rng, n_steps)

		rollout = rollout_mgr.reset()

		def _scan_rollout(scan_carry, rng):
			(rollout, 
			 state, 
			 obs, 
			 carry,
			 done, 
			 extra, 
			 ep_stats) = scan_carry
			
			next_scan_carry = \
				self._get_transition(
					rng,
					pop, 
					rollout_mgr,
					rollout,
					params, 
					state, 
					obs, 
					carry, 
					done,
					reset_state, 
					extra)

			(rollout, 
			 next_state, 
			 next_obs, 
			 next_carry, 
			 done, 
			 info, 
			 extra) = next_scan_carry

			if ep_stats is not None:
				_ep_stats_update_fn = self._update_ep_stats \
					if pop is self.student_pop else self._update_ued_ep_stats

				ep_stats = _ep_stats_update_fn(ep_stats, done, info)

			return (rollout, next_state, next_obs, next_carry, done, extra, ep_stats), None

		(rollout, state, obs, carry, done, extra, ep_stats), _ = jax.lax.scan(
			_scan_rollout,
			(rollout, state, obs, carry, done, extra, ep_stats),
			rngs,
			length=n_steps,
			unroll=self.n_unroll_rollout
		)

		return rollout, state, obs, carry, extra, ep_stats

	@partial(jax.jit, static_argnums=(0,))
	def _compile_stats(self, 
		update_stats, ep_stats, 
		ued_update_stats, ued_ep_stats,
		env_metrics=None,
		grad_stats=None, ued_grad_stats=None):
		mean_returns_by_student = jax.vmap(lambda x: x.mean())(ep_stats['return'])
		mean_returns_by_teacher = jax.vmap(lambda x: x.mean())(ued_ep_stats['return'])

		mean_ep_stats = jax.vmap(lambda info: jax.tree.map(lambda x: x.mean(), info))(
			{k:ep_stats[k] for k in self.rolling_stats.names}
		)
		ued_mean_ep_stats = jax.vmap(lambda info: jax.tree.map(lambda x: x.mean(), info))(
			{k:ued_ep_stats[k] for k in self.ued_rolling_stats.names}
		)

		student_stats = {
			f'mean_{k}':v for k,v in mean_ep_stats.items()
		}
		student_stats.update(update_stats)

		stats = {}
		for i in range(self.n_students):
			_student_stats = jax.tree_util.tree_map(lambda x: x[i], student_stats) # for agent0
			stats.update({f'{k}_a{i}':v for k,v in _student_stats.items()})

		teacher_stats = {
			f'mean_{k}_tch':v[0] for k,v in ued_mean_ep_stats.items()
		}
		teacher_stats.update({
			f'{k}_tch':v[0] for k,v in ued_update_stats.items()
		})
		stats.update(teacher_stats)

		if self.track_env_metrics:
			for k,v in env_metrics.items():
				if 'shortest_path' in k:
					env_metrics.update({k:jnp.max(v)})  # Track max shortest path length

			passable_mask = env_metrics.pop('passable')
			mean_env_metrics = jax.tree_util.tree_map(
				lambda x: (x*passable_mask).sum()/passable_mask.sum(), 
				env_metrics
			)
			mean_env_metrics.update({'passable_ratio': passable_mask.mean()})
			stats.update({
				f'env/{k}':v for k,v in mean_env_metrics.items()
			})

		if self.n_devices > 1:
			stats = jax.tree.map(lambda x: jax.lax.pmean(x, 'device'), stats)

		return stats

	def get_shmap_spec(self):
		runner_state_size = len(inspect.signature(self.run).parameters)
		in_spec = [P(None,'device'),]*(runner_state_size)
		out_spec = [P(None,'device'),]*(runner_state_size)

		in_spec[:2] = [P(None),]*2
		in_spec[6] = P(None)
		in_spec = tuple(in_spec)
		out_spec = (P(None),) + in_spec

		return in_spec, out_spec

	@partial(jax.jit, static_argnums=(0,))
	def run(
		self, 
		rng, 
		train_state, 
		state,
		obs,
		carry,
		ep_stats,
		ued_train_state,
		ued_state,
		ued_obs,
		ued_carry,
		ued_ep_stats):
		"""
		Perform one update step: rollout teacher + students
		"""
		if self.n_devices > 1:
			rng = jax.random.fold_in(rng, jax.lax.axis_index('device'))

		# === Reset teacher env + rollout teacher
		rng, *vrngs = jax.random.split(rng, self.teacher_pop.n_agents+1)
		ued_reset_out = self.benv.reset_teacher(jnp.array(vrngs))
		if len(ued_reset_out) > 2:
			ued_obs, ued_state, ued_extra = ued_reset_out
		else:
			ued_obs, ued_state = ued_reset_out
			ued_extra = None

		# Reset UED ep_stats
		if self.ued_rolling_stats is not None:
			ued_ep_stats = self.ued_rolling_stats.reset_stats(
				batch_shape=(1,self.n_parallel))
		else:
			ued_ep_stats = None

		tch_rollout_batch_shape = (1,self.n_parallel*self.n_eval)
		done = jnp.zeros(tch_rollout_batch_shape, dtype=jnp.bool_)
		rng, subrng = jax.random.split(rng)
		ued_rollout, ued_state, ued_obs, ued_carry, _, ued_ep_stats = \
			self._rollout(
				subrng,
				self.teacher_pop,
				self.teacher_rollout,
				self.teacher_n_rollout_steps,
				jax.lax.stop_gradient(ued_train_state.params), 
				ued_state, 
				ued_obs, 
				ued_carry,
				done, 
				extra=ued_extra, 
				ep_stats=ued_ep_stats
			)

		# === Reset student to new envs + rollout students
		rng, *vrngs = jax.random.split(rng, self.teacher_pop.n_agents+1)
		obs, state, extra = jax.tree_util.tree_map(
			lambda x:x.squeeze(0), self.benv.reset_student(
				jnp.array(vrngs),
				ued_state, 
				self.student_pop.n_agents))
		reset_state = state

		# Reset student ep_stats
		st_rollout_batch_shape = (self.n_students,self.n_parallel*self.n_eval)
		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=st_rollout_batch_shape)

		done = jnp.zeros(st_rollout_batch_shape, dtype=jnp.bool_)
		rng, subrng = jax.random.split(rng)
		rollout, state, obs, carry, extra, ep_stats = \
			self._rollout(
				subrng, 
				self.student_pop,
				self.student_rollout,
				self.n_rollout_steps,
				jax.lax.stop_gradient(train_state.params),
				state, 
				obs, 
				carry, 
				done,
				reset_state=reset_state, 
				extra=extra, 
				ep_stats=ep_stats)

		# === Update student with PPO
		# PPOAgent vmaps over the train state and batch. Batch must be N x EM
		student_rollout_last_value = self.student_pop.get_value(
			jax.lax.stop_gradient(train_state.params), obs, carry
		)
		train_batch = self.student_rollout.get_batch(
			rollout, 
			student_rollout_last_value
		)

		rng, subrng = jax.random.split(rng)
		train_state, update_stats = self.student_pop.update(subrng, train_state, train_batch)

		# === Update teacher with PPO
		# - Compute returns per env per agent
		# - Compute batched returns based on returns per env per agent
		ued_score, _ = compute_ued_scores(self.ued_score, train_batch, self.n_eval)
		ued_rollout = self.teacher_rollout.set_final_reward(ued_rollout, ued_score)
		ued_train_batch = self.teacher_rollout.get_batch(
			ued_rollout, 
			jnp.zeros((1, self.n_parallel)) # Last step terminates episode
		)

		ued_ep_stats = self._update_ued_ep_stats(
			ued_ep_stats, 
			jnp.ones((1,len(ued_score),1), dtype=jnp.bool_),
			{'return': jnp.expand_dims(ued_score, (0,-1))}
		)

		# Update teacher, batch must be 1 x Ex1
		rng, subrng = jax.random.split(rng)
		ued_train_state, ued_update_stats = self.teacher_pop.update(subrng, ued_train_state, ued_train_batch)

		# --------------------------------------------------
		# Collect metrics
		if self.track_env_metrics:
			env_metrics = self.benv.get_env_metrics(reset_state)
		else:
			env_metrics = None

		grad_stats, ued_grad_stats = None, None

		stats = self._compile_stats(
			update_stats, ep_stats, 
			ued_update_stats, ued_ep_stats,
			env_metrics,
			grad_stats, ued_grad_stats)
		stats.update(dict(n_updates=train_state.n_updates[0]))

		train_state = train_state.increment()
		ued_train_state = ued_train_state.increment()
		self.n_updates += 1

		return (
			stats, 
			rng,
			train_state, state, obs, carry, ep_stats,
			ued_train_state, ued_state, ued_obs, ued_carry, ued_ep_stats
		)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from enum import Enum
from collections import namedtuple

import numpy as np
import jax
import jax.numpy as jnp
import chex


class UEDScore(Enum):
	RELATIVE_REGRET = 1
	MEAN_RELATIVE_REGRET = 2
	POPULATION_REGRET = 3
	RETURN = 4
	NEG_RETURN = 5
	L1_VALUE_LOSS = 6
	POSITIVE_VALUE_LOSS = 7
	MAX_MC = 8
	VALUE_DISAGREEMENT = 9


@partial(jax.jit, static_argnums=(2,3))
def compute_episodic_stats(
	metrics, 
	dones, 
	time_average=False, 
	partial_metrics=0,
	partial_steps=0,
	return_partial=False):
	env_batch_shape = dones.shape[1:]
	n_episodes = jnp.zeros(env_batch_shape, dtype=jnp.uint32)
	sum_ep_metrics = jnp.zeros(env_batch_shape, dtype=jnp.float32)
	partial_metrics = jnp.zeros(env_batch_shape, dtype=jnp.float32)
	max_metrics = jnp.zeros(env_batch_shape, dtype=jnp.float32)
	steps = jnp.zeros(env_batch_shape, dtype=jnp.float32)
	partial_steps = jnp.zeros(env_batch_shape, dtype=jnp.float32)

	def _compute_metrics(carry, step):
		(n_episodes, 
		 sum_ep_metrics, 
		 max_metrics,
		 partial_metrics, 
		 partial_steps) = carry

		_metrics, _dones = step

		partial_metrics += _metrics
		partial_steps += 1

		if time_average:
			ep_metric = partial_metrics/partial_steps
		else:
			ep_metric = partial_metrics

		sum_ep_metrics += _dones*ep_metric
		max_metrics = _dones*jnp.maximum(max_metrics, ep_metric) + (1-_dones)*max_metrics

		n_episodes += _dones

		partial_metrics = (1-_dones)*partial_metrics
		partial_steps = (1-_dones)*partial_steps

		return (
			n_episodes, 
			sum_ep_metrics, 
			max_metrics,
			partial_metrics,
			partial_steps
		), None

	# @todo: Track and return partial means + partial n_steps
	(n_episodes, sum_ep_metrics, max_metrics, partial_metrics, partial_steps), _  = jax.lax.scan(
		_compute_metrics,
		(n_episodes, sum_ep_metrics, max_metrics, partial_metrics, partial_steps),
		(metrics, dones),
		length=len(metrics)
	)

	# Take mean over eval dimension
	total_metrics_per_env = sum_ep_metrics.sum(-1)
	n_episodes_per_env = n_episodes.sum(-1)
	n_episodes_per_env = jnp.maximum(n_episodes_per_env, 1)

	# Take max over eval dimension
	max_metrics_per_env = max_metrics.max(-1)

	return total_metrics_per_env/n_episodes_per_env, max_metrics_per_env

@partial(jax.jit, static_argnums=(0,))
def _compute_ued_scores(score_name: UEDScore, batch: namedtuple, info=None):
	"""
	Compute UED score from a rollout batch.
	Individual score functions return a tuple of mean_scores and max_scores,
	where each is of dimension n_agents x n_envs.
	"""
	if score_name in [UEDScore.RELATIVE_REGRET, UEDScore.MEAN_RELATIVE_REGRET, UEDScore.POPULATION_REGRET]:
		mean_scores, max_scores, score_info = compute_return(batch)

	elif score_name == UEDScore.RETURN:
		mean_scores, max_scores, score_info = compute_return(batch)

	elif score_name == UEDScore.NEG_RETURN:
		batch = batch._replace(rewards=-batch.rewards)
		mean_scores, max_scores, score_info = compute_return(batch)

	elif score_name == UEDScore.MAX_MC:
		mean_scores, max_scores, score_info = compute_max_mc(batch, info)

	elif score_name == UEDScore.L1_VALUE_LOSS:
		mean_scores, max_scores, score_info = compute_l1_value_loss(batch)

	elif score_name == UEDScore.POSITIVE_VALUE_LOSS:
		mean_scores, max_scores, score_info = compute_positive_value_loss(batch)

	elif score_name == UEDScore.VALUE_DISAGREEMENT:
		mean_scores, max_scores, score_info = compute_value_disagreement(batch)

	return mean_scores, max_scores, score_info

@partial(jax.jit, static_argnums=(0,2,4,5))
def compute_ued_scores(score_name: UEDScore, batch: namedtuple, n_eval: int, info: dict = None, ignore_val=None, per_agent=False):
	n_agents, n_steps, flat_batch_size = batch.dones.shape
	pop_batch_shape = (n_agents, n_steps, flat_batch_size//n_eval, n_eval)
	batch = jax.tree_util.tree_map(lambda x: x.reshape(*pop_batch_shape, *x.shape[3:]), batch)

	mean_env_returns_per_agent, max_env_returns_per_agent, score_info = \
		jax.vmap(_compute_ued_scores, in_axes=(None, 0, 0))(
			score_name, batch, info
		)

	if score_name in [UEDScore.RELATIVE_REGRET, UEDScore.MEAN_RELATIVE_REGRET]:
		assert len(mean_env_returns_per_agent) == 2, \
			"Standard PAIRED requires exactly 2 agents."		

	if score_name == UEDScore.RELATIVE_REGRET:
		scores = jnp.clip(max_env_returns_per_agent[1] \
				- mean_env_returns_per_agent[0], 0)

	elif score_name == UEDScore.MEAN_RELATIVE_REGRET:
		scores = jnp.clip(mean_env_returns_per_agent[1] \
				- mean_env_returns_per_agent[0], 0)

	elif score_name == UEDScore.POPULATION_REGRET:
		max_env_returns = max_env_returns_per_agent.max(0)
		mean_env_returns = mean_env_returns_per_agent.mean(0)
		scores = max_env_returns - mean_env_returns
	else:
		if per_agent:
			scores = mean_env_returns_per_agent
			max_scores = max_env_returns_per_agent
		else:
			scores = mean_env_returns_per_agent.mean(0)
			max_scores = max_env_returns_per_agent.max(0)

	if ignore_val is not None:
		if per_agent:
			incomplete_idxs = batch.dones.sum(axis=(1,-1)) == 0
		else:
			incomplete_idxs = batch.dones.sum(axis=(0, 1,-1)) == 0

		scores = jnp.where(incomplete_idxs, ignore_val, scores)

	return scores, score_info

# ======== UED score computations ========
def compute_return(batch):
	mean_scores, max_scores = compute_episodic_stats(
		batch.rewards, batch.dones, time_average=False)

	return mean_scores, max_scores, None


def compute_l1_value_loss(batch):
	mean_scores, max_scores = compute_episodic_stats(
		jnp.abs(batch.advantages), batch.dones, time_average=True)

	return mean_scores, max_scores, None


def compute_positive_value_loss(batch):
	mean_scores, max_scores = compute_episodic_stats(
		jnp.clip(batch.advantages, 0), batch.dones, time_average=True)

	return mean_scores, max_scores, None


def compute_max_mc(batch, info):
	_, max_env_returns_per_agent = \
		compute_episodic_stats(batch.rewards, batch.dones, time_average=False)

	max_returns = jnp.maximum(max_env_returns_per_agent, info['max_returns'])
	mean_scores, max_scores = compute_episodic_stats(
		max_returns[jnp.newaxis,:,jnp.newaxis] - batch.values, # Can be negative
		batch.dones, 
		time_average=True
	)

	score_info = {'max_returns': max_env_returns_per_agent}

	return mean_scores, max_scores, score_info


def compute_value_disagreement(batch):
	mean_scores, max_scores = compute_episodic_stats(
		batch.values.std(-1), batch.dones, time_average=True
	)

	return mean_scores, max_scores, None 




"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file is modified from
https://github.com/RobertTLange/gymnax/

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from typing import Tuple, Union
from collections import OrderedDict
import chex
import jax
import jax.numpy as jnp


class Discrete(object):
	"""
	Minimal jittable class for discrete gymnax spaces.
	TODO: For now this is a 1d space. Make composable for multi-discrete.
	"""

	def __init__(self, num_categories: int, shape: Tuple[int] = (), dtype=jnp.int32):
		assert num_categories >= 0
		self.n = num_categories
		self.shape = shape
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from set of categorical choices."""
		return jax.random.randint(
			rng, shape=self.shape, minval=0, maxval=self.n
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(x >= 0, x < self.n)
		return range_cond


class Box(object):
	"""
	Minimal jittable class for array-shaped gymnax spaces.
	TODO: Add unboundedness - sampling from other distributions, etc.
	"""
	def __init__(
		self,
		low: float,
		high: float,
		shape: Tuple[int],
		dtype: jnp.dtype = jnp.float32,
	):
		self.low = low
		self.high = high
		self.shape = shape
		self.dtype = dtype

	def sample(self, rng: chex.PRNGKey) -> chex.Array:
		"""Sample random action uniformly from 1D continuous range."""
		return jax.random.uniform(
			rng, shape=self.shape, minval=self.low, maxval=self.high
		).astype(self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether specific object is within space."""
		# type_cond = isinstance(x, self.dtype)
		# shape_cond = (x.shape == self.shape)
		range_cond = jnp.logical_and(
			jnp.all(x >= self.low), jnp.all(x <= self.high)
		)
		return range_cond


class Dict(object):
	"""Minimal jittable class for dictionary of simpler jittable spaces."""
	def __init__(self, spaces: dict):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> dict:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return OrderedDict(
			[
				(k, self.spaces[k].sample(key_split[i]))
				for i, k in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, dict)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for k, space in self.spaces.items():
			out_of_space += 1 - space.contains(getattr(x, k))
		return out_of_space == 0


class Tuple(object):
	"""Minimal jittable class for tuple (product) of jittable spaces."""
	def __init__(self, spaces: Union[tuple, list]):
		self.spaces = spaces
		self.num_spaces = len(spaces)

	def sample(self, rng: chex.PRNGKey) -> Tuple[chex.Array]:
		"""Sample random action from all subspaces."""
		key_split = jax.random.split(rng, self.num_spaces)
		return tuple(
			[
				space.sample(key_split[i])
				for i, space in enumerate(self.spaces)
			]
		)

	def contains(self, x: jnp.int_) -> bool:
		"""Check whether dimensions of object are within subspace."""
		# type_cond = isinstance(x, tuple)
		# num_space_cond = len(x) != len(self.spaces)
		# Check for each space individually
		out_of_space = 0
		for space in self.spaces:
			out_of_space += 1 - space.contains(x)
		return out_of_space == 0


class Dummy(object):
	def __init__(self, default_value=None):
		self._default_value = default_value
		self.dtype = jnp.uint32
		if self._default_value is None:
			self.n = 0
		else:
			self.n = 1

	def sample(self, rng: chex.PRNGKey):
		if self._default_value is None:
			return None
		else:
			return jnp.array(0, dtype=self.dtype)

	def contains(self, x: jnp.int_) -> bool:
		return False



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import chex
from typing import Tuple, Union, Optional

from minimax.envs.environment import EnvState


class EnvWrapper:
	"""
	Abstract base class for an env wrapper.
	"""
	def __init__(self, env):
		self._env = env

		self._wrap_level = 1
		while hasattr(env, '_env'):
			if isinstance(env, EnvWrapper):
				self._wrap_level += 1

			env = env._env

	@classmethod
	def is_compatible(cls, env):
		return True

	@property
	def base_env(self):
		env = self
		for i in range(self._wrap_level):
			env = env._env

		return env

	def reset_extra(self):
		return {}

	def get_monitored_metrics(self):
		if self._wrap_level > 1:
			return self._env.get_monitored_metrics()
		return ()

	def _append_extra_to_tuple(self, _tuple, extra=None):
		if extra is None:
			extra = self.reset_extra()
		
		if self._wrap_level > 1:
			_tuple[-1].update(extra)
		else:
			_tuple = _tuple + (extra,)

		return _tuple

	def step(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
		reset_state: Optional[chex.ArrayTree] = None,
		extra: dict = None,
	) -> Tuple[chex.Array, EnvState, float, bool]:
		if self._wrap_level > 1:
			return self._env.step(key, state, action, reset_state, extra)
		else:
			_tuple = self._env.step(
				key, state, action, reset_state=reset_state
			)
			return self._append_extra_to_tuple(_tuple, extra) 

	def reset(
		self, 
		key: chex.PRNGKey,
	) -> Tuple[chex.Array, EnvState, chex.ArrayTree]:
		_tuple = self._env.reset(key)

		return self._append_extra_to_tuple(_tuple)

	def set_state(
		self,
		state: EnvState,
	) -> Tuple[chex.ArrayTree, EnvState, chex.ArrayTree]:
		_tuple = self._env.set_state(state)

		return self._append_extra_to_tuple(_tuple)

	def set_env_instance(
		self,
		encoding: chex.ArrayTree,
	) -> Tuple[chex.ArrayTree, EnvState, chex.ArrayTree]:
		_tuple = self._env.set_env_instance(encoding)

		return self._append_extra_to_tuple(_tuple)

	def reset_student(
		self, 
		key: chex.PRNGKey,
		state: chex.ArrayTree,
	) -> Tuple[chex.ArrayTree, EnvState, chex.ArrayTree]:
		_tuple = self._env.reset_student(key, state)

		return self._append_extra_to_tuple(_tuple)

	def __getattr__(self, attr):
		return getattr(self._env, attr)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import pytest
import numpy as np
import jax
import jax.numpy as jnp

import tests.dummy_test_envs as dummy_test_envs
import envs
from envs.environment_ued import UEDEnvironment
from envs.wrappers import EnvWrapper
from envs.wrappers import UEDEnvWrapper


class TestEnvWrapper:
	def setup_class(self):
		# Set up environment + wrapper
		env_kwargs = dict(
			reward_per_step=1,
			max_episode_steps=3
		)
		env_kwargs = env_kwargs
		ued_env_kwargs = env_kwargs
		self.env_kwargs = env_kwargs

		self.env, self.env_params, self.ued_params = \
			envs.make(
				'DummyRewardEnv', 
				env_kwargs=env_kwargs,
				ued_env_kwargs=ued_env_kwargs,
				**self._get_wrappers()
			)

		self.dummy_rng = jax.random.PRNGKey(0)

	@staticmethod
	def _get_wrappers():
		return {
			'wrappers':['env_wrapper'],
			'ued_wrappers':['ued_env_wrapper']
		}


class TestDefaultEnvWrapper(TestEnvWrapper):
	def setup_class(self):
		# Set up environment + wrapper
		env_kwargs = dict(
			reward_per_step=1,
			max_episode_steps=2
		)
		env_kwargs = env_kwargs
		ued_env_kwargs = env_kwargs

		self.env, self.env_params, self.ued_params = \
			envs.make(
				'DummyRewardEnv', 
				env_kwargs=env_kwargs,
				ued_env_kwargs=ued_env_kwargs,
				**self._get_wrappers()
			)

		self.dummy_rng = jax.random.PRNGKey(0)

	def test_base_env(self):
		assert isinstance(self.env.base_env, UEDEnvironment)

	def test_reset_extra(self):
		extra = self.env.reset_extra()
		assert len(extra) == 0

	def test_step(self):
		obs, state, extra = self.env.reset(self.dummy_rng)
		extra = self.env.step(self.dummy_rng, state, 0)[-1]
		assert len(extra) == 0

	def test_reset(self):
		obs, state, extra = self.env.reset(self.dummy_rng)
		assert len(extra) == 0

	def test_reset_env_instance(self):
		_, ued_state = self.env.ued_env.reset(self.dummy_rng)
		instance = self.env.ued_env.get_env_instance(self.dummy_rng, ued_state)
		extra = self.env.set_env_instance(instance)[-1]
		assert len(extra) == 0

	def reset_teacher(self):
		out = self.env.reset_teacher(self.dummy_rng)
		assert len(out) == 2

	def step_teacher(self):
		_, ued_state = self.env.reset_teacher(self.dummy_rng)
		out = self.env.step_teacher(self.dummy_rng, ued_state, 0)
		assert len(out) == 5

	def reset_student(self):
		_, ued_state = self.env.reset_teacher(self.dummy_rng)
		_, state, extra = self.env.reset_student(self.dummy_rng, ued_state)
		assert len(extra) == 0


class TestMonitorReturnWrapper(TestEnvWrapper):
	@staticmethod
	def _get_wrappers():
		return {
			'wrappers':['monitor_return']
		}

	def test_wrap_level(self):
		assert self.env._wrap_level == 1

	def test_reset_extra(self):
		extra = self.env.reset_extra()
		assert len(extra) == 1 and extra['ep_return'] == 0

	def test_get_monitored_metrics(self):
		metrics = self.env.get_monitored_metrics()
		assert len(metrics) == 1 and 'return' in metrics

	def test_reset(self):
		_, _, extra = self.env.reset(self.dummy_rng)
		assert len(extra) == 1 and extra['ep_return'] == 0

	def test_step(self):
		obs, state, extra = self.env.reset(self.dummy_rng)

		n_steps = 2
		return_ = 0
		for i in range(n_steps):
			_, state, r, _, _, extra = self.env.step(
				self.dummy_rng, state, 0, extra=extra)
			return_ += r

		assert extra['ep_return'] == self.env_kwargs['reward_per_step']*n_steps

		# Finish the episode
		_, state, r, _, info, extra = self.env.step(
			self.dummy_rng, state, 0, extra=extra)

		assert extra['ep_return'] == 0
		assert info['return'] == self.env_kwargs['reward_per_step']*(n_steps+1)


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Union, Optional

from .env_wrapper import EnvWrapper
from minimax.envs.environment import EnvState


class MonitorEpisodicMetricsWrapper(EnvWrapper):
	"""
	Tracks episodic metrics about environment instances.
	"""
	def __init__(self, env):
		super().__init__(env)

		base_env = env.base_env if hasattr(env, 'base_env') else env
		_env = base_env.env if hasattr(base_env, 'env') else base_env

		self.metrics = ()
		if hasattr(_env, 'get_episodic_metrics'):
			reset_tuple = _env.reset(jax.random.PRNGKey(0))
			dummy_state = reset_tuple[1]

			self.metrics = tuple({
				k: jnp.zeros_like(v) \
					for k,v in _env.get_episodic_metrics(dummy_state).items()
			}.keys())

	@classmethod
	def is_compatible(cls, env):
		_env = env.env if hasattr(env, 'env') else env
		return hasattr(_env, 'get_episodic_metrics')

	def get_monitored_metrics(self):
		metrics = tuple(f'ep/{m}' for m in self.metrics)
		if self._wrap_level > 1:
			return self._env.get_monitored_metrics() + metrics
		else:
			return self.metrics

	def step(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
		reset_state: Optional[chex.ArrayTree] = None,
		extra: dict = None,
	) -> Tuple[chex.Array, EnvState, float, bool]:
		step_kwargs = dict(
			reset_state=reset_state
		)
		if self._wrap_level > 1:
			step_kwargs.update(dict(
				extra=extra
			))

		step = self._env.step(
				key, 
				state, 
				action, 
				**step_kwargs)

		if len(step) == 5:
			obs, state, reward, done, info = step
		else:
			obs, state, reward, done, info, extra = step

		if len(self.metrics) > 0:
			for m in self.metrics:
				info[f'ep/{m}'] = info[m]
				del info[m]

		return obs, state, reward, done, info, extra



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import importlib
import copy

# Global registry
registered_models = {}


def _load(name):
	mod_name, attr_name = name.split(":")
	mod = importlib.import_module(mod_name)
	fn = getattr(mod, attr_name)
	return fn


def _get_register_id(env_group_id, model_id):
	return f"{env_group_id.lower()}-{model_id}"


def register(env_group_id, model_id, entry_point):
	register_id = _get_register_id(env_group_id, model_id)
	if register_id in registered_models:
		raise ValueError(f'A model has already been registered as {register_id}.')
	else:
		registered_models[register_id] = entry_point


def make(
	env_name, model_name=None, **model_kwargs):
	env_group_id = env_name.split('-')[0].lstrip('UED')
	model_id = model_name

	register_id = _get_register_id(env_group_id, model_id)
	if register_id not in registered_models:
		raise ValueError(f'No model for {register_id} found.')
	else:
		entry = registered_models[register_id]

		if callable(entry):
			model = entry(**model_kwargs)
		else:
			model = _load(entry)(**model_kwargs)

		return model


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from typing import Any, Callable

import jax
import jax.numpy as jnp
from flax import core
from flax import struct
import optax
import chex

from .plr import PLRBuffer


class VmapTrainState(struct.PyTreeNode):
  n_iters: chex.Array
  n_updates: chex.Array # per agent
  n_grad_updates: chex.Array # per agent
  apply_fn: Callable = struct.field(pytree_node=False)
  params: core.FrozenDict[str, Any]
  tx: optax.GradientTransformation = struct.field(pytree_node=False)
  opt_state: optax.OptState
  plr_buffer: PLRBuffer = None

  def apply_gradients(self, *, grads, **kwargs):
    updates, new_opt_state = self.tx.update(
        grads, self.opt_state, self.params)
    new_params = optax.apply_updates(self.params, updates)

    return self.replace(
        n_grad_updates=self.n_updates + 1,
        params=new_params,
        opt_state=new_opt_state,
        **kwargs,
    )

  @classmethod
  def create(cls, *, 
  	apply_fn, 
  	params, 
  	tx,
  	**kwargs
  ):
    opt_state = jax.vmap(tx.init)(params)
    return cls(
        n_iters=jnp.array(jax.vmap(lambda x: 0)(params), dtype=jnp.uint32),
        n_updates=jnp.array(jax.vmap(lambda x: 0)(params), dtype=jnp.uint32),
        n_grad_updates=jnp.array(jax.vmap(lambda x: 0)(params), dtype=jnp.uint32),
        apply_fn=apply_fn,
        params=params,
        tx=tx,
        opt_state=opt_state,
        **kwargs,
    )

  def increment(self):
    return self.replace(
      n_iters=self.n_iters + 1,
    )

  def increment_updates(self):
    return self.replace(
      n_updates=self.n_updates + 1,
    ) 

  @property
  def state_dict(self):
    return dict(
      n_iters=self.n_iters,
      n_updates=self.n_updates,
      n_grad_updates=self.n_grad_updates,
      params=self.params,
      opt_state=self.opt_state
    )

  def load_state_dict(self, state):
    return self.replace(
      n_iters=state['n_iters'],
      n_updates=state['n_updates'],
      n_grad_updates=state['n_grad_updates'],
      params=state['params'],
      opt_state=state['opt_state']
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial


def _get_base_role(role):
	return role.removesuffix('_tch').removesuffix('_st')


def _get_runner_info(p):
	n_students = p.get('n_students', 1)
	n_eval = p.get('n_eval', 1)

	n_devices = p.get('n_devices', 1)
	device_info = ''
	if n_devices > 1:
		device_info = f'_d{n_devices}'

	return f"r{n_students}s_{p.n_parallel}p_{n_eval}e_{p.n_rollout_steps}t_ae{p.adam_eps}{device_info}"


def _get_runner_info_dr(p):
	ac_info = _get_runner_info(p)

	reset_info = ""
	if p.ac_reset_env_on_rollout:
		reset_info = f"r"
	if len(reset_info) > 0:
		reset_info = f"_{reset_info}"

	return f"{ac_info}{reset_info}"


def _get_ued_runner_info(p):
	info = _get_runner_info(p)

	if p.ued_score == 'relative_regret':
		ued_score = 'r'
	elif p.ued_score == 'mean_relative_regret':
		ued_score = 'mr'
	elif p.ued_score == 'population_regret':
		ued_score = 'p'
	elif p.ued_score == 'neg_return':
		ued_score = 'nr'
	elif p.ued_score == 'l1_value_loss':
		ued_score = 'l1v'
	elif p.ued_score == 'positive_value_loss':
		ued_score = 'pvl'
	elif p.ued_score == 'max_mc':
		ued_score = 'mm'
	elif p.ued_score == 'value_disagreement':
		ued_score = 'vd'
	else:
		raise ValueError(f'Unsupported ued_score {ued_score}')

	info = f"{info}_s{ued_score}"

	return info


def _get_plr_runner_info(p):
	info = _get_ued_runner_info(p)

	plr_info = f'p{p.plr_replay_prob}b{p.plr_buffer_size}t{p.plr_temp}s{p.plr_staleness_coef}m{p.plr_min_fill_ratio}'
	if p.plr_use_score_ranks:
		plr_info = f'{plr_info}r'

	if p.plr_mutation_fn:
		plr_info = f'{plr_info}_m{p.plr_mutation_fn[:3]}{p.plr_n_mutations}{p.plr_mutation_criterion[:3]}'
		if p.plr_mutation_criterion != "batch":
			plr_info = f"{plr_info}{p.plr_mutation_subsample_size}"

	plr_prefix = ''
	if p.plr_use_robust_plr:
		plr_prefix += 'r'
	if p.plr_use_parallel_eval:
		plr_prefix += 'p'
	if p.plr_force_unique:
		plr_prefix += 'f'
	if len(plr_prefix) > 0:
		plr_prefix += '_'

	return f"{plr_prefix}{plr_info}_{info}"


def _get_runner_info_paired(p):
	return _get_ued_runner_info(p)


def _get_env_info_default(p):
	return p.env_name.lower().replace('-', '_')


def _get_env_info_maze(p):
	see_agent = 'na' if not p.maze_see_agent else ''

	placement_info = ""
	if p.maze_replace_wall_pos:
		placement_info = f'f'
	if p.maze_sample_n_walls:
		placement_info = f"{placement_info}s"
	if len(placement_info) > 0:
		placement_info = f"_{placement_info}"

	return f"{p.env_name}{p.maze_height}x{p.maze_width}w{p.maze_n_walls}{see_agent}{placement_info}"


def _get_env_info_maze_ued(p):
	see_agent = 'na' if not p.maze_see_agent else ''

	info = f"_{see_agent}_ld{p.maze_ued_noise_dim}"

	placement_info = ""
	if p.maze_ued_fixed_n_wall_steps:
		placement_info = f"f"
	if p.maze_ued_replace_wall_pos:
		placement_info = f"{placement_info}r"
	if p.maze_ued_set_agent_dir:
		placement_info = f"{placement_info}d"
	if p.maze_ued_first_wall_pos_sets_budget:
		placement_info = f"{placement_info}b"
	if len(placement_info) > 0:
		placement_info = f"_{placement_info}"
	info = f"{info}{placement_info}"

	return f"{p.env_name}{p.maze_height}x{p.maze_width}w{p.maze_n_walls}{info}"


def _get_model_info_maze_default(p, role):
	model_info = ''
	if f'{role}_recurrent_arch' in p and p[f'{role}_recurrent_arch'] is not None:
		model_info = f"{p[f'{role}_recurrent_arch']}_h{p[f'{role}_recurrent_hidden_dim']}"

		if p[f'{role}_recurrent_arch'] == 's5':
			model_info = f"{model_info}nb{p.get(f'{role}_s5_n_blocks', 1)}nl{p.get(f'{role}_s5_n_layers',4)}"

			activation = p.get(f'{role}_s5_activation')
			if activation == 'half_glu1':
				activation = 'hg1'
			elif activation == 'half_glu2':
				activation = 'hg2'
			elif activation == 'full_glu':
				activation = 'fg'
			elif activation == 'gelu':
				activation = 'g'
			else:
				activation = 'hg1'
			model_info = f'a{activation}_{model_info}'

			ln_key =  f'{role}_s5_layernorm_pos'
			ln_info = None
			if ln_key in p:
				ln = p[ln_key]
				if ln == 'pre':
					ln_info = 'pr'
				elif ln == 'post':
					ln_info = 'po'

			if ln_info is not None:
				model_info = f"l{ln_info}_{model_info}"


	model_info = f'_{model_info}' if len(model_info) > 0 else ''

	value_info = ''
	value_ensemble_key = f'{role}_value_ensemble_size'
	value_ensemble_size = p.get(value_ensemble_key)
	if value_ensemble_size and value_ensemble_size > 1:
		value_info = f've{value_ensemble_size}'

	base_activation = p.get(f'{role}_base_activation', 'relu')[:2]

	model_info = f"h{p[f'{role}_hidden_dim']}cf{p[f'{role}_n_conv_filters']}fc{p[f'{role}_n_hidden_layers']}se{p[f'{role}_scalar_embed_dim']}ba_{base_activation}{model_info}{value_info}"

	return model_info


def _get_algo_info_ppo(p, role):
	if role == 'student':
		lr = str(p.lr)
		if 'lr_final' in p:
			lr_final = '' if p.lr_final is None or p.lr_final == p.lr else str(p.lr_final)
			if len(lr_final) > 0:
				lr = f"{lr}_{lr_final}"

		return f"ppo_lr{lr}g{p.discount}cv{p.student_value_loss_coef}ce{p.student_entropy_coef}e{p.student_ppo_n_epochs}mb{p.student_ppo_n_minibatches}l{p.student_gae_lambda}_pc{p.student_ppo_clip_eps}"
	else:
		if 'teacher_lr' in p:
			teacher_lr = str(p.lr) if p.teacher_lr is None else str(p.teacher_lr)
		else:
			teacher_lr = str(p.lr)

		if 'teacher_lr_final' in p:
			teacher_lr_final = str(p.lr_final) if p.teacher_lr_final is None else str(p.teacher_lr_final)
		else:
			teacher_lr_final = str(p.lr_final) if 'lr_final' in p else ''

		if teacher_lr_final == teacher_lr:
			teacher_lr_final = ''

		if len(teacher_lr_final) > 0:
			teacher_lr = f"{teacher_lr}_{teacher_lr_final}"

		return f"ppo_lr{teacher_lr}g{p.teacher_discount}cv{p.teacher_value_loss_coef}ce{p.teacher_entropy_coef}e{p.teacher_ppo_n_epochs}mb{p.teacher_ppo_n_minibatches}l{p.teacher_gae_lambda}pc{p.teacher_ppo_clip_eps}"


# ============================================================

RUNNER_INFO_HANDLERS = {
	'dr': _get_runner_info_dr,
	'plr': _get_plr_runner_info,
	'paired': _get_runner_info_paired,
}

ENV_INFO_HANDLERS = {
	'maze': _get_env_info_maze,
	'maze_ued': _get_env_info_maze_ued,
}

MODEL_INFO_HANDLERS = {
	'maze': {
		'default_student_cnn': partial(_get_model_info_maze_default, role='student'),
		'default_teacher_cnn': partial(_get_model_info_maze_default, role='teacher'),
	},
}

ALGO_INFO_HANDLERS = {
	'ppo': _get_algo_info_ppo
}


def get_runner_info(p):
	return RUNNER_INFO_HANDLERS[p.get('train_runner', 'dr')](p)

def get_env_info(p):
	p.env_name = p.env_name.lower()
	env_name = p.env_name
	if p.train_runner in ['paired',]:
		env_name = f'{env_name}_ued'

	return ENV_INFO_HANDLERS.get(
		env_name, _get_env_info_default
	)(p)

def get_model_info(p, role='student'):
	model_name = p.get(f'{role}_model_name')
	if model_name is None:
		model_name = p['student_model_name']
	env_name = p.env_name.lower().split('-')[0]

	return MODEL_INFO_HANDLERS[env_name][model_name](p)

def get_algo_info(p, role='student'):
	return ALGO_INFO_HANDLERS[p.agent_rl_algo](p, role)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .maze import Maze, UEDMaze
from .batch_env import BatchEnv
from .batch_env_ued import BatchUEDEnv

from .registration import make, get_comparator, get_mutator


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from enum import IntEnum

import numpy as np
import jax
import jax.numpy as jnp

from .common import make_maze_map
from minimax.envs.registration import register_mutator


class Mutations(IntEnum):
    # Turn left, turn right, move forward
    NO_OP = 0
    FLIP_WALL = 1
    MOVE_GOAL = 2


def flip_wall(rng, state):
	wall_map = state.wall_map
	h,w = wall_map.shape
	wall_mask = jnp.ones((h*w,), dtype=jnp.bool_)

	goal_idx = w*state.goal_pos[1] + state.goal_pos[0]
	agent_idx = w*state.agent_pos[1] + state.agent_pos[0]
	wall_mask = wall_mask.at[goal_idx].set(False)
	wall_mask = wall_mask.at[agent_idx].set(False)

	flip_idx = jax.random.choice(rng, np.arange(h*w), p=wall_mask)
	flip_y = flip_idx//w
	flip_x = flip_idx%w

	flip_val = ~wall_map.at[flip_y,flip_x].get()
	next_wall_map = wall_map.at[flip_y,flip_x].set(flip_val)

	return state.replace(wall_map=next_wall_map)


def move_goal(rng, state):
	wall_map = state.wall_map
	h,w = wall_map.shape
	wall_mask = wall_map.flatten()

	goal_idx = w*state.goal_pos[1] + state.goal_pos[0]
	agent_idx = w*state.agent_pos[1] + state.agent_pos[0]
	wall_mask = wall_mask.at[goal_idx].set(True)
	wall_mask = wall_mask.at[agent_idx].set(True)

	next_goal_idx = jax.random.choice(rng, np.arange(h*w), p=~wall_mask)
	next_goal_y = next_goal_idx//w
	next_goal_x = next_goal_idx%w

	next_wall_map = wall_map.at[next_goal_y,next_goal_x].set(False)
	next_goal_pos = jnp.array([next_goal_x,next_goal_y], dtype=jnp.uint32)

	return state.replace(wall_map=next_wall_map, goal_pos=next_goal_pos)


@partial(jax.jit, static_argnums=(1,3))
def move_goal_flip_walls(rng, params, state, n=1):
	if n == 0:
		return state

	def _mutate(carry, step):
		state = carry
		rng, mutation = step

		rng, arng, brng = jax.random.split(rng,3)

		is_flip_wall = jnp.equal(mutation, Mutations.FLIP_WALL.value)
		mutated_state = flip_wall(arng, state)
		next_state = jax.tree_map(lambda x,y: jax.lax.select(is_flip_wall, x, y), mutated_state, state)

		is_move_goal = jnp.equal(mutation, Mutations.MOVE_GOAL.value)
		mutated_state = move_goal(brng, state)
		next_state = jax.tree_map(lambda x,y: jax.lax.select(is_move_goal, x, y), mutated_state, next_state)

		return next_state, None

	rng, nrng, *mrngs = jax.random.split(rng, n+2)
	mutations = jax.random.choice(nrng, np.arange(len(Mutations)), (n,))

	state, _ = jax.lax.scan(_mutate, state, (jnp.array(mrngs), mutations))

	# Update state maze_map
	next_maze_map = make_maze_map(
		params,
		state.wall_map, 
		state.goal_pos, 
		state.agent_pos, 
		state.agent_dir_idx, 
		pad_obs=True)

	return state.replace(maze_map=next_maze_map)

# Register the mutators
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register_mutator(env_id='Maze', mutator_id=None, entry_point=module_path + ':move_goal_flip_walls')


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp

import minimax.envs as envs


class BatchUEDEnv:
	"""
	Wraps and batches a UEDEnvironment in 
	its private methods as follows:
	
	For student MDP:
		Manages a batch of n_parallel x n_eval envs

	For teacher MDP:
		Manages a batch of n_parallel envs.

	The public interface vmaps the private methods over
	an additional agent population dimension.
	"""
	def __init__(
		self,
		env_name,
		n_parallel,
		n_eval,
		env_kwargs,
		ued_env_kwargs,
		wrappers=['monitor_return'],
		ued_wrappers=None):
		self.env, self.env_params, self.ued_params = \
			envs.make(
				env_name, 
				env_kwargs=env_kwargs,
				ued_env_kwargs=ued_env_kwargs,
				wrappers=wrappers,
				ued_wrappers=ued_wrappers
			)

		self.n_parallel = n_parallel
		self.n_eval = n_eval
		self.sub_batch_size = n_parallel*n_eval

		self.reset_student = jax.vmap(self._reset_student, in_axes=(0,0,None))
		self.step_teacher = jax.vmap(self._step_teacher, in_axes=0)
		self.step_student = jax.vmap(self._step_student, in_axes=0)
		
		self.set_env_instance = jax.vmap(self._set_env_instance, in_axes=0)
		self.get_env_metrics = jax.vmap(self._get_env_metrics, in_axes=0)

	partial(jax.jit, static_argnums=(2,))
	def reset(self, rng, sub_batch_size=None):
		if sub_batch_size is None:
			sub_batch_size = self.sub_batch_size

		return jax.vmap(self._reset, in_axes=(0,None))(rng, sub_batch_size)

	def _reset(self, rng, sub_batch_size):
		brngs = jax.random.split(rng, sub_batch_size)
		return jax.vmap(self.env.reset)(brngs)

	partial(jax.jit, static_argnums=(2,))
	def reset_teacher(self, rng, n_parallel=None):
		if n_parallel is None:
			n_parallel = self.n_parallel

		return jax.vmap(self._reset_teacher, in_axes=(0,None))(rng, n_parallel)

	def _reset_teacher(self, rng, n_parallel):
		"""
		Reset n_parallel envs
		"""
		brngs = jax.random.split(rng, n_parallel)
		return jax.vmap(self.env.reset_teacher)(brngs)

	def _step_teacher(self, rng, ued_state, action, extra=None):
		"""
		Step n_parallel envs
		"""
		brngs = jax.random.split(rng, self.n_parallel)
		step_args = (brngs, ued_state, action)
		if extra is not None:
			step_args += (extra,)

		return jax.vmap(self.env.step_teacher)(*step_args)

	def _reset_student(self, rng, ued_state, n_students):
		"""
		Reset the student MDP based on the state of the teacher MDP.
		"""
		brngs = jax.random.split(rng, self.n_parallel)
		obs, state, extra = \
			jax.vmap(self.env.reset_student)(brngs, ued_state)

		obs = jax.tree_util.tree_map(
			lambda x: jnp.repeat(
						jnp.expand_dims(jnp.repeat(x, self.n_eval, 0), 0), n_students, 0), obs)

		state = jax.tree_util.tree_map(
			lambda x: jnp.repeat(
				jnp.expand_dims(jnp.repeat(x, self.n_eval, 0), 0), n_students, 0), state)

		extra = jax.tree_util.tree_map(
			lambda x: jnp.repeat(
				jnp.expand_dims(jnp.repeat(x, self.n_eval, 0), 0), n_students, 0), extra)

		return obs, state, extra

	def _step_student(self, rng, state, action, reset_state, extra=None):
		"""
		Step the student MDP.
		"""
		brngs = jax.random.split(rng, self.sub_batch_size)
		return jax.vmap(self.env.step)(brngs, state, action, reset_state, extra)

	def _set_env_instance(self, instance):
		"""
		Reset the student MDP to a particular configuration,
		captured by state argument. Used for PLR.
		"""
		return jax.vmap(self.env.set_env_instance)(instance)

	def _get_env_metrics(self, state):
		return jax.vmap(self.env.get_env_metrics)(state) 



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from dataclasses import dataclass
from collections import namedtuple, OrderedDict
from functools import partial
from enum import IntEnum

import numpy as np
import jax
import jax.numpy as jnp
from jax import lax
from typing import Tuple, Optional
import chex
from flax import struct
from flax.core.frozen_dict import FrozenDict

from minimax.envs import environment, spaces
from minimax.envs.registration import register
import minimax.util.graph as _graph_util
from .common import (
	OBJECT_TO_INDEX,
	COLORS,
	COLOR_TO_INDEX,
	DIR_TO_VEC,
	EnvInstance, 
	make_maze_map)


class Actions(IntEnum):
    # Turn left, turn right, move forward
    left = 0
    right = 1
    forward = 2

    # Pick up an object
    pickup = 3
    # Drop an object
    drop = 4
    # Toggle/activate an object
    toggle = 5

    # Done completing task
    done = 6


@struct.dataclass
class EnvState:
	agent_pos: chex.Array
	agent_dir: chex.Array
	agent_dir_idx: int
	goal_pos: chex.Array
	wall_map: chex.Array
	maze_map: chex.Array
	time: int
	terminal: bool


@struct.dataclass
class EnvParams:
	height: int = 15
	width: int = 15
	n_walls: int = 25 
	agent_view_size: int = 5
	replace_wall_pos: bool = False
	see_through_walls: bool = True
	see_agent: bool = False
	normalize_obs: bool = False
	sample_n_walls: bool = False # Sample n_walls uniformly in [0, n_walls]
	obs_agent_pos: bool = False
	max_episode_steps: int = 250
	singleton_seed: int = -1,


class Maze(environment.Environment):
	def __init__(
		self,
		height=13,
		width=13,
		n_walls=25,
		agent_view_size=5,
		replace_wall_pos=False,
		see_through_walls=True,
		see_agent=False,
		max_episode_steps=250,
		normalize_obs=False,
		sample_n_walls=False,
		obs_agent_pos=False,
		singleton_seed=-1
	):
		super().__init__()

		self.obs_shape = (agent_view_size, agent_view_size, 3)
		self.action_set = jnp.array([
			Actions.left,
			Actions.right,
			Actions.forward,
			Actions.pickup,
			Actions.drop,
			Actions.toggle,
			Actions.done
		])

		self.params = EnvParams(
			height=height,
			width=width,
			n_walls=n_walls,
			agent_view_size=agent_view_size,
			replace_wall_pos=replace_wall_pos and not sample_n_walls,
			see_through_walls=see_through_walls,
			see_agent=see_agent,
			max_episode_steps=max_episode_steps,
			normalize_obs=normalize_obs,
			sample_n_walls=sample_n_walls,
			obs_agent_pos=obs_agent_pos,
			singleton_seed=-1,
		)

	@property
	def default_params(self) -> EnvParams:
		# Default environment parameters
		return EnvParams()

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: int,
	) -> Tuple[chex.Array, EnvState, float, bool, dict]:
		"""Perform single timestep state transition."""
		a = self.action_set[action]
		state, reward = self.step_agent(key, state, a)
		# Check game condition & no. steps for termination condition
		state = state.replace(time=state.time + 1)
		done = self.is_terminal(state)
		state = state.replace(terminal=done)

		return (
			lax.stop_gradient(self.get_obs(state)),
			lax.stop_gradient(state),
			reward.astype(jnp.float32),
			done,
			{},
		)

	def reset_env(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.Array, EnvState]:
		"""Reset environment state by resampling contents of maze_map
		- initial agent position
		- goal position
		- wall positions
		"""
		params = self.params
		h = params.height
		w = params.width
		all_pos = np.arange(np.prod([h, w]), dtype=jnp.uint32)

		# Reset wall map, with shape H x W, and value of 1 at (i,j) iff there is a wall at (i,j)
		key, subkey = jax.random.split(key)
		wall_idx = jax.random.choice(
			subkey, all_pos, 
			shape=(params.n_walls,), 
			replace=params.replace_wall_pos)

		if params.sample_n_walls:
			key, subkey = jax.random.split(key)
			sampled_n_walls = jax.random.randint(subkey, (), minval=0, maxval=params.n_walls)
			sample_wall_mask = jnp.arange(params.n_walls) < sampled_n_walls
			dummy_wall_idx = wall_idx.at[0].get().repeat(params.n_walls)
			wall_idx = jax.lax.select(
				sample_wall_mask,
				wall_idx,
				dummy_wall_idx
			)

		occupied_mask = jnp.zeros_like(all_pos)
		occupied_mask = occupied_mask.at[wall_idx].set(1)
		wall_map = occupied_mask.reshape(h, w).astype(jnp.bool_)

		# Reset agent position + dir
		key, subkey = jax.random.split(key)
		agent_idx = jax.random.choice(subkey, all_pos, shape=(1,), p=(~occupied_mask.astype(jnp.bool_)).astype(jnp.float32))
		occupied_mask = occupied_mask.at[agent_idx].set(1)
		agent_pos = jnp.array([agent_idx%w,agent_idx//w], dtype=jnp.uint32).flatten()

		key, subkey = jax.random.split(key)
		agent_dir_idx = jax.random.choice(subkey, jnp.arange(len(DIR_TO_VEC), dtype=jnp.uint8))
		agent_dir = DIR_TO_VEC.at[agent_dir_idx].get()

		# Reset goal position
		key, subkey = jax.random.split(key)
		goal_idx = jax.random.choice(subkey, all_pos, shape=(1,), p=(~occupied_mask.astype(jnp.bool_)).astype(jnp.float32))
		goal_pos = jnp.array([goal_idx%w,goal_idx//w], dtype=jnp.uint32).flatten()

		maze_map = make_maze_map(
			params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx, 
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=agent_dir,
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map.astype(jnp.bool_),
			maze_map=maze_map,
			time=0,
			terminal=False,
		)

		return self.get_obs(state), state

	def set_env_instance(
			self,  
			encoding: EnvInstance):
		"""
		Instance is encoded as a PyTree containing the following fields:
		agent_pos, agent_dir, goal_pos, wall_map
		"""
		params = self.params
		agent_pos = encoding.agent_pos
		agent_dir_idx = encoding.agent_dir_idx

		agent_dir = DIR_TO_VEC.at[agent_dir_idx].get()
		goal_pos = encoding.goal_pos
		wall_map = encoding.wall_map
		maze_map = make_maze_map(
			params,
			wall_map, 
			goal_pos, 
			agent_pos, 
			agent_dir_idx, # ued instances include wall padding
			pad_obs=True)

		state = EnvState(
			agent_pos=agent_pos,
			agent_dir=agent_dir,
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map,
			maze_map=maze_map,
			time=0,
			terminal=False
		)

		return self.get_obs(state), state

	def get_obs(self, state: EnvState) -> chex.Array:
		"""Return limited grid view ahead of agent."""
		obs = jnp.zeros(self.obs_shape, dtype=jnp.uint8)
		
		agent_x, agent_y = state.agent_pos

		obs_fwd_bound1 = state.agent_pos
		obs_fwd_bound2 = state.agent_pos + state.agent_dir*(self.obs_shape[0]-1)

		side_offset = self.obs_shape[0]//2
		obs_side_bound1 = state.agent_pos + (state.agent_dir == 0)*side_offset
		obs_side_bound2 = state.agent_pos - (state.agent_dir == 0)*side_offset

		all_bounds = jnp.stack([obs_fwd_bound1, obs_fwd_bound2, obs_side_bound1, obs_side_bound2])

		# Clip obs to grid bounds appropriately
		padding = obs.shape[0]-1
		obs_bounds_min = np.min(all_bounds, 0) + padding
		obs_range_x = jnp.arange(obs.shape[0]) + obs_bounds_min[1]
		obs_range_y = jnp.arange(obs.shape[0]) + obs_bounds_min[0]

		meshgrid = jnp.meshgrid(obs_range_y,obs_range_x)
		coord_y = meshgrid[1].flatten()
		coord_x = meshgrid[0].flatten()

		obs = state.maze_map.at[
			coord_y,coord_x,:].get().reshape(obs.shape[0],obs.shape[1],3)

		obs = (state.agent_dir_idx == 0)*jnp.rot90(obs, 1) + \
			  	(state.agent_dir_idx == 1)*jnp.rot90(obs, 2) + \
			  	(state.agent_dir_idx == 2)*jnp.rot90(obs, 3) + \
			  	(state.agent_dir_idx == 3)*jnp.rot90(obs, 4)

		if not self.params.see_agent:
			obs = obs.at[-1, side_offset].set(
				jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
			)

		if not self.params.see_through_walls:
			pass

		image = obs.astype(jnp.uint8)
		if self.params.normalize_obs:
			image = image/10.0

		obs_dict = dict(
			image=image,
			agent_dir=state.agent_dir_idx
		)
		if self.params.obs_agent_pos:
			obs_dict.update(dict(agent_pos=state.agent_pos))

		return OrderedDict(obs_dict)

	def step_agent(self, key: chex.PRNGKey, state: EnvState, action: int) -> Tuple[EnvState, float]:
		params = self.params

		# Update agent position (forward action)
		fwd_pos = jnp.minimum(
			jnp.maximum(state.agent_pos + (action == Actions.forward)*state.agent_dir, 0), 
			jnp.array((params.width-1, params.height-1), dtype=jnp.uint32))

		# Can't go past wall or goal
		fwd_pos_has_wall = state.wall_map.at[fwd_pos[1], fwd_pos[0]].get()
		fwd_pos_has_goal = jnp.logical_and(fwd_pos[0] == state.goal_pos[0], fwd_pos[1] == state.goal_pos[1])

		fwd_pos_blocked = jnp.logical_or(fwd_pos_has_wall, fwd_pos_has_goal)

		agent_pos_prev = jnp.array(state.agent_pos)
		agent_pos = (fwd_pos_blocked*state.agent_pos + (~fwd_pos_blocked)*fwd_pos).astype(jnp.uint32)

		# Update agent direction (left_turn or right_turn action)
		agent_dir_offset = \
			0 \
			+ (action == Actions.left)*(-1) \
			+ (action == Actions.right)*1

		agent_dir_idx = (state.agent_dir_idx + agent_dir_offset) % 4
		agent_dir = DIR_TO_VEC[agent_dir_idx]

		# Update agent component in maze_map
		empty = jnp.array([OBJECT_TO_INDEX['empty'], 0, 0], dtype=jnp.uint8)
		agent = jnp.array([OBJECT_TO_INDEX['agent'], COLOR_TO_INDEX['red'], agent_dir_idx], dtype=jnp.uint8)
		padding = self.obs_shape[0]-1
		maze_map = state.maze_map
		maze_map = maze_map.at[padding+agent_pos_prev[1],padding+agent_pos_prev[0],:].set(empty)
		maze_map = maze_map.at[padding+agent_pos[1],padding+agent_pos[0],:].set(agent)

		# Return reward
		# rng = jax.random.PRNGKey(agent_dir_idx + agent_pos[0] + agent_pos[1])
		# rand_reward = jax.random.uniform(rng)
		reward = (1.0 - 0.9*((state.time+1)/params.max_episode_steps))*fwd_pos_has_goal # rand_reward

		return (
			state.replace(
				agent_pos=agent_pos,
				agent_dir_idx=agent_dir_idx,
				agent_dir=agent_dir,
				maze_map=maze_map,	
				terminal=fwd_pos_has_goal),
			reward
		)

	def is_terminal(self, state: EnvState) -> bool:
		"""Check whether state is terminal."""
		done_steps = state.time >= self.params.max_episode_steps
		return jnp.logical_or(done_steps, state.terminal)

	def get_eval_solved_rate_fn(self):
		def _fn(ep_stats):
			return ep_stats['return'] > 0

		return _fn

	@property
	def name(self) -> str:
		"""Environment name."""
		return "Maze"

	@property
	def num_actions(self) -> int:
		"""Number of actions possible in environment."""
		return len(self.action_set)

	def action_space(self) -> spaces.Discrete:
		"""Action space of the environment."""
		return spaces.Discrete(
			len(self.action_set),
			dtype=jnp.uint32
		)

	def observation_space(self) -> spaces.Dict:
		"""Observation space of the environment."""
		spaces_dict = {
			'image':spaces.Box(0, 255, self.obs_shape),
			'agent_dir': spaces.Discrete(4)
		}
		if self.params.obs_agent_pos:
			params = self.params
			h = params.height
			w = params.width
			spaces_dict.update({'agent_pos': spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32)})

		return spaces.Dict(spaces_dict)

	def state_space(self) -> spaces.Dict:
		"""State space of the environment."""
		params = self.params
		h = params.height
		w = params.width
		agent_view_size = params.agent_view_size
		return spaces.Dict({
			"agent_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
			"agent_dir": spaces.Discrete(4),
			"goal_pos": spaces.Box(0, max(w, h), (2,), dtype=jnp.uint32),
			"maze_map": spaces.Box(0, 255, (w + agent_view_size, h + agent_view_size, 3), dtype=jnp.uint32),
			"time": spaces.Discrete(params.max_episode_steps),
			"terminal": spaces.Discrete(2),
		})

	def max_episode_steps(self) -> int:
		return self.params.max_episode_steps

	def get_env_metrics(self, state: EnvState) -> dict:
		n_walls = state.wall_map.sum()
		shortest_path_length = _graph_util.shortest_path_len(
			state.wall_map,
			state.agent_pos,
			state.goal_pos
		)

		return dict(
			n_walls=n_walls,
			shortest_path_length=shortest_path_length,
			passable=shortest_path_length > 0,
		)


# Register the env
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register(env_id='Maze', entry_point=module_path + ':Maze')


if __name__ == '__main__':
	from envs.wrappers import MonitorReturnWrapper

	render = True

	if render:
		from envs.viz.grid_viz import GridVisualizer
		viz = GridVisualizer()
		obs_viz = GridVisualizer()

		viz.show()
		obs_viz.show()

	kwargs = dict(
		max_episode_steps=250,
		height=15,
		width=15,
		n_walls=25,
		agent_view_size=5,
		see_through_walls=True
	)
	env = MonitorReturnWrapper(Maze(**kwargs))
	params = env.params
	extra = env.reset_extra()

	jit_reset_env = jax.jit(env.reset)
	jit_step_env = jax.jit(env.step)

	key, subkey = jax.random.split(jax.random.PRNGKey(0))
	obs, state, extra = jit_reset_env(subkey)

	import time
	for i in range(1000):
		print('step', i)
		key, subkey = jax.random.split(key)
		start = time.time()
		obs, state, reward, done, info, extra = jit_step_env(
			key, state, action=env.action_space().sample(subkey), extra=extra)
		obs['image'].block_until_ready()
		end = time.time()
		print(f"sps: {1/(end-start)}")
		print('return:', info['return'])

		if render:
			viz.render(params, state)
			obs_viz.render_grid(np.asarray(env.get_obs(state)['image']), k_rot90=0, agent_dir_idx=3)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np


@partial(jax.jit, static_argnums=(1,))
def apsp(A, n=None):
	"""
	Compute APSP for adjacency matrix A 
	using Seidel's algorithm.
	"""
	if n is None:
		n = A.shape[0]
	assert(n == A.shape[0]), 'n must equal dim of A.'

	n_steps = int(np.ceil(np.log(n)/np.log(2)))
	A_cache = jnp.zeros((n_steps, n, n), dtype=jnp.uint32)
	steps_to_reduce = jnp.array(1, dtype=jnp.int32)

	def _scan_fwd_step(carry, step): 
		i = step
		A, A_cache, steps_to_reduce = carry
		A_cache = A_cache.at[i].set(A)

		Z = A@A
		B = jnp.logical_or(
				A == 1, 
				Z > 0
			).astype(jnp.uint32) \
			 .at[jnp.diag_indices(n)].set(0)
		A = B

		complete = B.sum() - jnp.diagonal(B).sum() == n*(n-1)
		steps_to_reduce += ~complete

		return (A, A_cache, steps_to_reduce), None

	(B, A_cache, steps_to_reduce), _ = jax.lax.scan(
		_scan_fwd_step,
		(A, A_cache, 1),
		jnp.arange(n_steps),
		length=n_steps
	)

	D = 2*B - A_cache[steps_to_reduce-1]

	def _scan_bkwd_step(carry, step):
		i = step
		(T, A_cache,steps_to_reduce) = carry

		A = A_cache[steps_to_reduce - i - 1]
		X = T@A

		thresh = T*(jnp.tile(A.sum(0, keepdims=True), (n, 1)))
		D = 2*T*(X >= thresh) + (2*T - 1)*(X < thresh)
		T = D*(i < steps_to_reduce) + T*(i >= steps_to_reduce)

		return (T, A_cache, steps_to_reduce), None

	(D, _, _), _ = jax.lax.scan(
		_scan_bkwd_step,
		(D, A_cache, steps_to_reduce),
		jnp.arange(1, n_steps),
		length=n_steps-1
	)

	return D


@jax.jit
def grid_to_graph(grid):
	"""
	Transform a binary grid (True == wall) into a
	graph, where walls are all connected to a default node.
	"""
	h, w = grid.shape
	nodes = grid.flatten()
	n = len(nodes)
	A = jnp.zeros((n,n), dtype=jnp.uint32)
	
	all_idx = jnp.arange(n)
	dum_neighbor_idx = jnp.argmax(~nodes)
	dum_neighbor_mask = jnp.zeros(n, dtype=jnp.bool_)
	dum_neighbor_mask = \
		dum_neighbor_mask.at[dum_neighbor_idx].set(True)

	def _get_neigbors(idx):
		# Return length n boolean mask of neighbors
		# We then vmap this function over all n
		r = idx + 1
		l = idx - 1
		t = idx - w
		b = idx + w

		l_border = jnp.logical_or(
			idx % w == 0,
			nodes[l]
		)
		r_border = jnp.logical_or(
			r % w == 0,
			nodes[r]
		)
		t_border = jnp.logical_or(
			idx // w == 0,
			nodes[t],
		)
		b_border = jnp.logical_or(
			idx // w == h - 1,
			nodes[b]
		)

		l_ignore = jnp.logical_or(
			l_border,
			nodes[idx]
		)
		r_ignore = jnp.logical_or(
			r_border,
			nodes[idx]
		)
		t_ignore = jnp.logical_or(
			t_border,
			nodes[idx]
		)
		b_ignore = jnp.logical_or(
			b_border,
			nodes[idx]
		)

		left = l*(1-l_ignore) + idx*(l_ignore)
		right = r*(1-r_ignore) + idx*(r_ignore)
		top = t*(1-t_ignore) + idx*(t_ignore)
		bottom = b*(1-b_ignore) + idx*(b_ignore)

		neighbor_mask = jnp.zeros(n, dtype=jnp.bool_)
		neighbor_mask = neighbor_mask.at[jnp.array([left, right, top, bottom])].set(True)

		neighbor_mask = (1-nodes[idx])*neighbor_mask + nodes[idx]*dum_neighbor_mask

		neighbor_mask = neighbor_mask.at[idx].set(False)

		return neighbor_mask

	A = jax.vmap(_get_neigbors)(all_idx).astype(dtype=jnp.uint32)
	A = jnp.maximum(A, A.transpose())

	return A
	

NEIGHBOR_OFFSETS = jnp.array([
	[1,0], # right
	[0,1], # bottom
	[-1,0], # left
	[0,-1], # top
	[0,0] # self
], dtype=jnp.int32)


@jax.jit
def component_mask_with_pos(grid, pos_a):
	"""
	Return a mask set to True in all cells that are
	a part of the connected component containing pos_a.
	"""
	h,w = grid.shape
	visited_mask = grid

	pos = pos_a
	visited_mask = visited_mask.at[
		pos[1],pos[0]
	].set(True)
	vstack = jnp.zeros((h*w, 2), dtype=jnp.uint32)
	vstack = vstack.at[:2].set(pos)
	vstack_size = 2

	def _scan_dfs(carry, step):
		(visited_mask, vstack, vstack_size) = carry

		pos = vstack[vstack_size-1]

		neighbors = \
			jnp.minimum(
				jnp.maximum(
					pos + NEIGHBOR_OFFSETS, 0
				), jnp.array([[h,w]])
			).astype(jnp.uint32)

		neighbors_mask = visited_mask.at[
			neighbors[:,1],neighbors[:,0]
		].get()
		n_neighbor_visited = neighbors_mask[:4].sum()
		all_visited = n_neighbor_visited == 4
		all_visited_post = n_neighbor_visited >= 3
		neighbors_mask = neighbors_mask.at[-1].set(~all_visited)

		next_neighbor_idx = jnp.argmax(~neighbors_mask)
		next_neighbor = neighbors[next_neighbor_idx]

		visited_mask = visited_mask.at[
			next_neighbor[1],next_neighbor[0]
		].set(True)

		vstack_size -= all_visited_post

		vstack = vstack.at[vstack_size].set(next_neighbor)
		vstack_size += ~all_visited

		pos = next_neighbor

		return (visited_mask, vstack, vstack_size), None

	max_n_steps = 2*h*w
	(visited_mask, vstack, vstack_size), _ = jax.lax.scan(
		_scan_dfs,
		(visited_mask, vstack, vstack_size),
		jnp.arange(max_n_steps),
		length=max_n_steps
	)

	visited_mask = visited_mask ^ grid
	return visited_mask


@jax.jit
def shortest_path_len(grid, pos_a, pos_b):
	grid = ~component_mask_with_pos(grid, pos_a)

	A = grid_to_graph(grid)
	D = apsp(A, n=A.shape[0])

	if len(pos_b.shape) == 2: # batch eval
		return jax.vmap(_shortest_path_len, in_axes=(None, None, 0, None))(
			grid, pos_a, pos_b, D
		)
	else:
		return _shortest_path_len(grid, pos_a, pos_b, D)


@jax.jit
def _shortest_path_len(grid, pos_a, pos_b, D):
	h,w = grid.shape

	a_idx = pos_a[1]*w + pos_a[0]
	b_idx = pos_b[1]*w + pos_b[0]
	d = D[a_idx][b_idx]

	mhttn_d = jnp.sum(jnp.abs(jnp.maximum(pos_a,pos_b)- jnp.minimum(pos_a,pos_b)))

	impossible = jnp.logical_and(
		d == 1,
		mhttn_d > 1
	)

	return d*(1-impossible)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from collections import defaultdict
import argparse
import sys
import re
import pprint

from minimax.util import DefaultDotDict, DotDict


def append_subparser_prefix(prefix, func):
	def prefixed_add_argument(*args, **kwargs):
		if len(args) > 0:
			name = args[0]
			if name.startswith('--'):
				name = f'--{prefix}_{name[2:]}'
			args = (name,) + args[1:]

		return func(*args, **kwargs)

	return prefixed_add_argument


def ensure_args_suffix(name):
	if not name.endswith('args'):
		name = f'{name}_args'

	return name


def get_all_cmd_arg_names():
	cmd_args = [s.removeprefix('--') for s in sys.argv if s.startswith('--')]
	arg_names = [x.split('=')[0] for x in cmd_args]

	return set(arg_names)


def get_argument_kwargs(subparser):
	kwargs = []
	skip_list = ['-h', '--help']
	for k, info in subparser.__dict__['_option_string_actions'].items():
		if k in skip_list:
			continue

		info_dict = info.__dict__
		kwargs_dict = dict(
			# option_strings=info_dict['option_strings'],
			name=info_dict['dest'],
			const=info_dict['const'],
			default=info_dict['default'],
			type=info_dict['type'],
			choices=info_dict['choices'],
			required=info_dict['required']
		)

		nargs = info_dict['nargs']
		if nargs == '?' or (nargs is not None and int(nargs) > 0):
			kwargs_dict.update(dict(nargs=info_dict['nargs']))

		kwargs.append(kwargs_dict)

	return kwargs


class Parsnip:
	"""
	Wraps a collection of argparse instances
	to enable convenient grouping of arguments and
	access via a DotDict-style interface.
	"""
	def __init__(self, description=None):
		self._base_parser = \
			argparse.ArgumentParser(description=description)
		self._subparsers = {}
		self._prefixes = []
		self._dependencies = defaultdict()
		self._dests = defaultdict()
		self._dependent_args = set()

	def add_subparser(
		self, 
		name, 
		prefix=None,
		dest=None,
		depends_on=None,
		dependency=None,
		is_individual_arg=False,
		description=None
	):	
		if not is_individual_arg:
			name = ensure_args_suffix(name)

		assert name not in self._subparsers, \
			f'Multiple subparsers named {name} detected.'

		if dependency is not None:
			if depends_on is not None:
				depends_on = ensure_args_suffix(depends_on)

				assert depends_on in self._subparsers, \
					f'Missing subparse {depends_on} must be added before dependent {name}.'

			assert isinstance(dependency, dict), \
				f'Subparser dependencies must be specified as dicts.'

			self._dependencies[name] = (depends_on, dependency)

		if dest is not None:
			dest = ensure_args_suffix(dest)
			assert dest in self._subparsers, \
				f"Missing dest {dest} must be specified before source {name}."

		subparser = argparse.ArgumentParser(
			description=description,
		 	allow_abbrev=False)
		if prefix is not None:
			subparser.add_argument = append_subparser_prefix(
				prefix, subparser.add_argument, 
			)

		self._subparsers[name] = subparser
		self._prefixes.append(prefix)
		self._dests[name] = dest
		
		return subparser

	def add_dependent_argument(
		self,
		*args,
		**kwargs,):

		assert 'dependency' in kwargs, \
			'Must specify dependency in kwargs.'
		dependency = kwargs.pop('dependency')

		name = args[0].removeprefix('--')

		prefix = kwargs.pop('prefix', None)
		dest = kwargs.pop('dest', None)

		subparser = self.add_subparser(
			name, 
			prefix=prefix,
			dependency=dependency,
			dest=dest,
			is_individual_arg=True,
			description=kwargs.pop('description', '')
		)
		subparser.add_argument(*args, **kwargs)

		self._dependent_args.add(name)

	def copy_arguments(
		self,
		src,
		dest,
		arg_prefix=None
	):
		all_subparsers = [k for k in self._subparsers.keys()]
		src_name = f'{src}_args'
		src_subparser = self._subparsers[src_name]
		src_idx = all_subparsers.index(src_name)
		src_prefix = self._prefixes[src_idx]

		dest_name = f'{dest}_args'
		dest_subparser = self._subparsers[dest_name]
		dest_idx = all_subparsers.index(dest_name)

		arg_prefix = f'{arg_prefix}_' if arg_prefix else ''

		argument_kwargs = get_argument_kwargs(src_subparser)
		for kwargs in argument_kwargs:
			name = kwargs.pop('name')
			flag = f"--{arg_prefix}{name.removeprefix(f'{src_prefix}_')}"

			dest_subparser.add_argument(
				flag,
				**kwargs
			)

	def parse_args(self, preview=False):
		cmd_arg_names, args, arg_data, _ = self._parse_cmd_line_flags()

		for name in cmd_arg_names:
			if name not in arg_data:
				raise ValueError(f'Unknown argument {name}.')

		if preview:
			print('🥕 Parsnip harvested the following arguments:')
			pp = pprint.PrettyPrinter(indent=4)
			pp.pprint(args)

		return args

	def parse_cmd_line_flags(self, as_grid_json=False):
		_, _, arg_data, _ = self._parse_cmd_line_flags()

		if as_grid_json:
			arg_data = {k:[v] for k,v in arg_data.items() if v is not None}

		return arg_data

	def _parse_cmd_line_flags(self, keep_structure=False):
		cmd_arg_names = get_all_cmd_arg_names()

		arg_data = {}
		argname2keypath = {}

		args = DefaultDotDict(vars(self._base_parser.parse_known_args()[0]))
		for k in args:
			arg_data[k] = args[k]
			argname2keypath[k] = [k]

		for i, (name, subparser) in enumerate(self._subparsers.items()):
			if name in self._dependencies:
				depends_on, dependencies = self._dependencies[name]
				if depends_on is None:
					dsubargs = arg_data
				else:
					dsubargs = arg_data[depends_on]
				skip_subparser = False
				for dk, dv in dependencies.items():
					if not isinstance(dv, (tuple, list)):
						dv = [dv]

					for v in dv:
						if isinstance(v, str):
							match_regex = f"^{v.replace('*', '.*')}$"
							match = re.match(match_regex, dsubargs[dk]) is not None
						else:
							match = dsubargs[dk] == v
						if match:
							break

					if not match:
						skip_subparser = True
						break

				if skip_subparser:
					continue

			prefix = self._prefixes[i]
			subargs, _ = subparser.parse_known_args()

			subargs = vars(subargs)
			for k in subargs:
				arg_data[k] = subargs[k]

			subargs = DotDict(
				{k.removeprefix(f'{prefix}_'):v for k,v in subargs.items()}
			)

			# Check for flatten and merge conditions
			dest = self._dests.get(name)
			if name in self._dependent_args:
				if dest is None:
					args[name] = subargs[name]
				else:
					args[dest].update({name: subargs[name]})
			else:
				if dest is None:
					args[name] = subargs
				else:
					args[dest].update(subargs)

			for k in subargs:
				if prefix is not None:
					argname = f'{prefix}_{k}'
				else:
					argname = k
				if dest is None:
					sp_name = name
				else:
					sp_name = dest
				if sp_name == k:
					argname2keypath[argname] = [k]
				else:
					argname2keypath[argname] = [sp_name, k]

		return cmd_arg_names, args, arg_data, argname2keypath

	@property
	def argname2keypath(self):
		args = DefaultDotDict(vars(self._base_parser.parse_known_args()[0]))
		argname2keypath = {}
		for k in args:
			argname2keypath[k] = args[k]

		for i, (name, subparser) in enumerate(self._subparsers.items()):
			if name in self._dependencies:
				depends_on, dependencies = self._dependencies[name]
				if depends_on is None:
					dsubargs = arg_data
				else:
					dsubargs = arg_data[depends_on]
				skip_subparser = False
				for dk, dv in dependencies.items():
					if not isinstance(dv, (tuple, list)):
						dv = [dv]

					for v in dv:
						if isinstance(v, str):
							match_regex = f"^{v.replace('*', '.*')}$"
							match = re.match(match_regex, dsubargs[dk]) is not None
						else:
							match = dsubargs[dk] == v
						if match:
							break

					if not match:
						skip_subparser = True
						break

				if skip_subparser:
					continue

			prefix = self._prefixes[i]
			subargs, _ = subparser.parse_known_args()

	def __getattr__(self, attr):
		# Default missing attr to _base_argparse instance
		return getattr(self._base_parser, attr)


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import pytest
import numpy as np
import jax
import jax.numpy as jnp

from util.rl import RolloutStorage
import envs
import models
import agents


class RequiresRolloutStorageTestClass:
	def setup_class(self):
		# Create maze object
		env_name = 'Maze'
		self.env, self.env_params = envs.make(env_name)

		self.n_steps = 32
		self.n_agents = 2
		self.n_envs = 3
		self.n_eval = 5
		self.rnn_dim = 4
		self.discount = 0.995
		self.gae_lambda = 0.95

		self.batch_shape = (
			self.n_agents, 
			self.n_steps, 
			self.n_envs*self.n_eval,
		)

		self.t_batch_shape = (
			self.n_agents, 
			self.n_envs*self.n_eval,
		)

		# Create dummy agent
		self.agent_model = models.make(
			env_name=env_name,
			model_name='default_student_cnn',
			recurrent_arch='lstm',
			recurrent_hidden_dim=self.rnn_dim
		)

		self.agent = agents.PPOAgent(
			model=self.agent_model,
		)

		dummy_rng = jax.random.PRNGKey(0)
		self.zero_carry_t = \
			self.agent.init_carry(
				dummy_rng, 
				batch_dims=(
					self.n_agents, 
					self.n_envs*self.n_eval)
			)

		# Initialize RolloutStorage obj
		self.rollout_mgr = RolloutStorage(
			discount=self.discount,
			gae_lambda=self.gae_lambda,
			n_envs=self.n_envs,
			n_eval=self.n_eval,
			n_steps=self.n_steps,
			action_space=self.env.action_space(),
			obs_space=self.env.observation_space(),
			agent=self.agent,
			n_agents=self.n_agents
		)

	def setup_method(self):
		pass

	def _get_dummy_step_components(self):
		t_batch_shape = self.t_batch_shape

		dummy_rng = jax.random.PRNGKey(0)
		obs, state = self.env.reset(dummy_rng)
		obs = jax.tree_util.tree_map(lambda x: x.repeat(
				np.prod(t_batch_shape)
			).reshape(
				*t_batch_shape, *x.shape
			), obs
		)
		action = jnp.ones(
			(*t_batch_shape, *self.env.action_space().shape), 
			dtype=self.env.action_space().dtype)

		done = jnp.ones(
			(*t_batch_shape, *self.env.action_space().shape), dtype=jnp.uint8)

		reward = jnp.ones(t_batch_shape, dtype=jnp.float32)

		log_pis_old = jnp.ones(t_batch_shape, dtype=jnp.float32)
		
		values_old = jnp.ones(t_batch_shape, dtype=jnp.float32)

		carry = self.zero_carry_t

		return (
			obs,
			action,
			reward,
			done,
			log_pis_old,
			values_old,
			carry
		)


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .training import VmapTrainState
from .agent_pop import AgentPop
from .rolling_stats import RollingStats
from .rollout_storage import RolloutStorage
from .ued_scores import *
from .plr import PLRManager, PopPLRManager